/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions

#include "quakedef.h"

// OK, so you don't get confused...
// vertex.attrib[0]			current position
// vertex.attrib[3]			previous position
// vertex.attrib[1]			texcoords
// vertex.attrib[2].x		current dot
// vertex.attrib[2].y		previous dot
// vertex.attrib[2].z		blendweight
// local 0					shadelight
// local 1					depth range OR blendweight
// local 2					texture scale
// local 3					fullbright scale
// local 4					shadevector
#define LERPPOSITION(blend) \
	"TEMP lerp;\n" \
	"SUB lerp, vertex.attrib[0], vertex.attrib[3];\n" \
	"MAD lerp, lerp, "blend", vertex.attrib[3];\n" \
	TRANSFORMVERTEX ("result.position", "lerp")

#define LERPPOSITIONVM(blend) \
	"TEMP lerp, xform;\n" \
	"SUB lerp, vertex.attrib[0], vertex.attrib[3];\n" \
	"MAD lerp, lerp, "blend", vertex.attrib[3];\n" \
	TRANSFORMVERTEX ("xform", "lerp") \
	"MUL result.position, xform, program.local[1];\n"

#define TRANSFORMTEXCOORD \
	"MUL result.texcoord[0], vertex.attrib[1], program.local[2];\n" \
	"MUL result.texcoord[1], vertex.attrib[1], program.local[3];\n"

#define LERPNORMAL(blend) \
	"TEMP shadedot, dothigh, dotlow;\n" \
	"SUB lerp, vertex.attrib[4], vertex.attrib[5];\n" \
	"MAD lerp, lerp, "blend", vertex.attrib[4];\n" \
	"DP3 shadedot, lerp, program.local[4];\n" \
	"ADD dothigh, shadedot, 1.0;\n" \
	"MAD dotlow, shadedot, 0.2954545, 1.0;\n" \
	"MAX shadedot, dotlow, dothigh;\n" \
	"MUL result.texcoord[2], shadedot, program.local[0];\n"

#define alias_vp \
	"!!ARBvp1.0\n" \
	LERPPOSITION ("program.local[1].x") \
	TRANSFORMTEXCOORD \
	LERPNORMAL ("program.local[1].x") \
	"END\n"

#define alias_vm_vp \
	"!!ARBvp1.0\n" \
	LERPPOSITIONVM ("vertex.attrib[2].x") \
	TRANSFORMTEXCOORD \
	LERPNORMAL ("vertex.attrib[2].x") \
	"END\n"

#define alias_fp \
	"!!ARBfp1.0\n" \
	"TEMP tex1;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	"MUL result.color, tex1, fragment.texcoord[2];\n" \
	"END\n"

#define aliasfb_fp \
	"!!ARBfp1.0\n" \
	"TEMP tex1, tex2;\n" \
	"TEX tex1, fragment.texcoord[0], texture[0], 2D;\n" \
	"TEX tex2, fragment.texcoord[0], texture[1], 2D;\n" \
	"MUL tex1, tex1, fragment.texcoord[2];\n" \
	"MAX result.color, tex1, tex2;\n" \
	"MOV result.color.a, fragment.texcoord[2].w; # otherwise we don't get alpha\n" \
	"END\n"


GLuint arb_alias_vp = 0;
GLuint arb_alias_vm_vp = 0;
GLuint arb_alias_fp = 0;
GLuint arb_aliasfb_fp = 0;
GLuint arb_alias_fp_fog = 0;
GLuint arb_aliasfb_fp_fog = 0;


#define shadow_vp \
	"!!ARBvp1.0\n" \
	LERPPOSITION ("program.local[1].x") \
	"END\n"

#define shadow_vm_vp \
	"!!ARBvp1.0\n" \
	LERPPOSITIONVM ("vertex.attrib[2].x") \
	"END\n"

#define shadow_fp \
	"!!ARBfp1.0\n" \
	"MOV result.color, program.local[0];\n" \
	"END\n"

GLuint arb_shadow_vp = 0;
GLuint arb_shadow_vm_vp = 0;
GLuint arb_shadow_fp = 0;
GLuint arb_shadow_fp_fog = 0;


void Mesh_CreateShaders (void)
{
	arb_alias_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, alias_vp, false);
	arb_alias_vm_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, alias_vm_vp, false);
	arb_alias_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, alias_fp, false);
	arb_aliasfb_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, aliasfb_fp, false);

	arb_shadow_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, shadow_vp, false);
	arb_shadow_vm_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, shadow_vm_vp, false);
	arb_shadow_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, shadow_fp, false);

	arb_alias_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, alias_fp, true);
	arb_aliasfb_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, aliasfb_fp, true);
	arb_shadow_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, shadow_fp, true);
}


GLuint r_meshvbo;
GLuint r_meshindexesvbo;

extern cvar_t gl_fullbrights;
extern gltexture_t *playertextures[];

extern vec3_t lightspot;
void R_MinimumLighting (float *lightcolor, float minimum);

// we'll reproduce the dotproduct from the actual correct normal
#define NUMVERTEXNORMALS	162

float r_avertexnormals[NUMVERTEXNORMALS][3] =
{
#include "anorms.h"
};

unsigned int r_meshindexbuffer = 0;
unsigned int r_meshvertexbuffer = 0;

void GL_MakeAliasModelDisplayLists (stvert_t *stverts, dtriangle_t *triangles)
{
	int i, j;

	// there can never be more than this number of verts and we just put them all on the hunk
	aliasmesh_t *desc = (aliasmesh_t *) Hunk_Alloc (sizeof (aliasmesh_t) * pheader->numverts);

	// there will always be this number of indexes
	unsigned short *indexes = (unsigned short *) Hunk_Alloc (sizeof (unsigned short) * pheader->numtris * 3);

	pheader->indexes = (intptr_t) indexes - (intptr_t) pheader;
	pheader->meshdesc = (intptr_t) desc - (intptr_t) pheader;
	pheader->numindexes = 0;
	pheader->numverts = 0;

	// NOTE
	// because not all skins in a model may be replaced by external textures we need to adjust texcoords for padding
	// per-skin instead of globally for the model.  see R_PadAliasTMU and R_UnpadAliasTMU in r_alias.c

	for (i = 0; i < pheader->numtris; i++)
	{
		for (j = 0; j < 3; j++)
		{
			int v;

			// index into hdr->vertexes
			unsigned short vertindex = triangles[i].vertindex[j];

			// basic s/t coords
			int s = stverts[vertindex].s;
			int t = stverts[vertindex].t;

			// check for back side and adjust texcoord s
			if (!triangles[i].facesfront && stverts[vertindex].onseam) s += pheader->skinwidth / 2;

			// see does this vert already exist
			for (v = 0; v < pheader->numverts; v++)
			{
				// it could use the same xyz but have different s and t
				if (desc[v].vertindex == vertindex && (int) desc[v].st[0] == s && (int) desc[v].st[1] == t)
				{
					// exists; emit an index for it
					indexes[pheader->numindexes++] = v;

					// no need to check any more
					break;
				}
			}

			if (v == pheader->numverts)
			{
				// doesn't exist; emit a new vert and index
				indexes[pheader->numindexes++] = pheader->numverts;

				desc[pheader->numverts].vertindex = vertindex;
				desc[pheader->numverts].st[0] = s;
				desc[pheader->numverts++].st[1] = t;
			}
		}
	}

	// create a hunk buffer for the final mesh we'll actually use
	{
		// vertex/fragment programs interpolate on the GPU and need to access the data directly
		unsigned short *hunkndx = (unsigned short *) Hunk_Alloc (sizeof (unsigned short) * pheader->numverts);
		qboolean *hunklerp = (qboolean *) Hunk_Alloc (sizeof (qboolean) * pheader->numverts);

		// move these to aliasmesh struct???
		pheader->vertindexes = (intptr_t) hunkndx - (intptr_t) pheader;
		pheader->lerptable = (intptr_t) hunklerp - (intptr_t) pheader;

		for (i = 0; i < pheader->numverts; i++)
		{
			// see note above for new handling of skin padding
			hunklerp[i] = true;
			hunkndx[i] = desc[i].vertindex;
		}

		pheader->delerped = false;
	}
}


void GLARB_SetupAliasModel (entity_t *e)
{
	// cull it (the viewmodel is never culled)
	if (r_shadows.value > 0.01f)
	{
		if (e != &cl.viewent)
		{
			if (R_CullBox (e->mins, e->maxs))
				e->visframe = -1;
			else e->visframe = r_framecount;
		}
	}
	else
	{
		if (e != &cl.viewent)
			if (R_CullBox (e->mins, e->maxs)) return;

		// the ent is visible now
		e->visframe = r_framecount;
	}
}


void GLARB_SetupAliasLighting (entity_t *e, aliasstate_t *state)
{
	float an;

	// get lighting information
	R_LightPoint (e, state->shadelight);
	VectorScale (state->shadelight, 2.0f, state->shadelight);
	VectorCopy (lightspot, e->aliasstate.lightspot);

	// the light scale will be doubled so halve the minimum
	if (e == &cl.viewent) R_MinimumLighting (state->shadelight, 36.0f);
	if (e->entnum >= 1 && e->entnum <= cl.maxclients) R_MinimumLighting (state->shadelight, 12.0f);

	// take to final scale - alpha will be stored in shadelight[3] separately
	VectorScale (state->shadelight, 1.0f / 255.0f, state->shadelight);

	// and now set up alpha
	state->shadelight[3] = (float) e->alpha / 255.0f;

	// other model types are assumed to provide their own normals or to use flat lighting
	if (e->model->type != mod_alias) return;

	// eval and store shadevector
	an = (state->angles[1] + state->angles[0]) / 180 * M_PI;
	state->shadevector[0] = cos (-an);
	state->shadevector[1] = sin (-an);
	state->shadevector[2] = 1;
	VectorNormalize (state->shadevector);
}


trivertx_t *GLARB_GetFrameTriVertx (aliashdr_t *hdr, int frame)
{
	// this prevents some code later on from making my eyes bleed
	return (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * frame);
}

void ScaleVerts (aliashdr_t *hdr, byte *original, float *scaled)
{
	scaled[0] = (float) original[0] * hdr->scale[0] + hdr->scale_origin[0];
	scaled[1] = (float) original[1] * hdr->scale[1] + hdr->scale_origin[1];
	scaled[2] = (float) original[2] * hdr->scale[2] + hdr->scale_origin[2];
}


void GLARB_DelerpViewModel (entity_t *e, aliashdr_t *hdr)
{
	int j;
	vec3_t	scaledf0, scaledf1;	// scaled versions of the verts (need to prescale for comparison)
	float	vdiff;				// difference in front to back movement
	trivertx_t	*vertsf0;
	trivertx_t	*vertsfi;

	qboolean *lerptable = (qboolean *) ((byte *) hdr + hdr->lerptable);

	if (hdr->nummeshframes < 2) return;
	if (hdr->delerped) return;

	vertsf0 = GLARB_GetFrameTriVertx (hdr, 0);
	vertsfi = GLARB_GetFrameTriVertx (hdr, 1);

	for (j = 0; j < hdr->vertsperframe; j++)
	{
		ScaleVerts (hdr, vertsf0->v, scaledf0);
		ScaleVerts (hdr, vertsfi->v, scaledf1);

		// get difference in front to back movement
		vdiff = scaledf1[0] - scaledf0[0];

		// if it's above a certain treshold, assume a muzzleflash and mark for nodraw
		// 10 is the approx lowest range of visible front to back in a view model, so that seems reasonable to work with
		if (vdiff > 10)
			lerptable[j] = false;
		else lerptable[j] = true;

		// next set of verts
		vertsf0++;
		vertsfi++;
	}

	hdr->delerped = true;
}


void GLARB_SetupAliasLerp (entity_t *e, aliashdr_t *hdr)
{
	int frame, posenum, numposes;

	if (e == &cl.viewent && !hdr->delerped)
	{
		GLARB_DelerpViewModel (e, hdr);
		hdr->delerped = true;
	}

	frame = e->frame;

	if ((frame >= hdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	posenum = hdr->frames[frame].firstpose;
	numposes = hdr->frames[frame].numposes;

	if (numposes > 1)
	{
		e->lerptime = hdr->frames[frame].interval;
		posenum += (int) (cl.time / e->lerptime) % numposes;
	}
	else e->lerptime = 0.1;

	if (e->lerpflags & LERP_RESETANIM)
	{
		// kill any lerp in progress
		e->lerpstart = 0;
		e->previouspose = posenum;
		e->currentpose = posenum;
		e->lerpflags &= ~LERP_RESETANIM;
	}
	else if (e->currentpose != posenum)
	{
		// definitely lerp now
		e->lerpstart = cl.time;
		e->previouspose = e->currentpose;
		e->currentpose = posenum;
	}

	if (e->previouspose == e->currentpose && e->currentpose == 0 && e != &cl.viewent)
	{
		e->lerpstart = 0;
		e->previouspose = posenum;
		e->currentpose = posenum;
		e->lerpflags &= ~LERP_RESETANIM;
	}
	else if (posenum == 0 && e == &cl.viewent)
	{
		e->lerpstart = 0;
		e->previouspose = posenum;
		e->currentpose = posenum;
		e->lerpflags &= ~LERP_RESETANIM;
	}

	e->aliasstate.blend = (cl.time - e->lerpstart) / e->lerptime;

	if (e->aliasstate.blend < 0) e->aliasstate.blend = 0;
	if (e->aliasstate.blend > 1) e->aliasstate.blend = 1;

	if (e->currentpose < e->previouspose)
	{
		// swap the poses and invert the blend for better cache locality
		e->aliasstate.currpose = e->previouspose;
		e->aliasstate.prevpose = e->currentpose;
		e->aliasstate.blend = 1.0f - e->aliasstate.blend;
	}
	else
	{
		e->aliasstate.currpose = e->currentpose;
		e->aliasstate.prevpose = e->previouspose;
	}

	// encode the final poses for use as a sort key
	e->encodedpose = e->aliasstate.prevpose | (e->aliasstate.currpose << 16);
}


void GLARB_LerpAliasMove (entity_t *e, aliasstate_t *state)
{
	int i;
	float blend;
	float d[3];

	if (e->lerpflags & LERP_RESETMOVE)
	{
		// kill any lerps in progress
		e->movelerpstart = 0;

		VectorCopy (e->origin, e->previousorigin);
		VectorCopy (e->origin, e->currentorigin);
		VectorCopy (e->angles, e->previousangles);
		VectorCopy (e->angles, e->currentangles);

		e->lerpflags &= ~LERP_RESETMOVE;
	}
	else if (!VectorCompare (e->origin, e->currentorigin) || !VectorCompare (e->angles, e->currentangles))
	{
		// origin/angles changed, start new lerp
		e->movelerpstart = cl.time;

		VectorCopy (e->currentorigin, e->previousorigin);
		VectorCopy (e->origin,  e->currentorigin);
		VectorCopy (e->currentangles, e->previousangles);
		VectorCopy (e->angles,  e->currentangles);
	}

	// set up values
	if (e != &cl.viewent && (e->lerpflags & LERP_MOVESTEP))
	{
		blend = (cl.time - e->movelerpstart) / 0.1;

		if (blend < 0) blend = 0;
		if (blend > 1) blend = 1;

		// translation
		VectorSubtract (e->currentorigin, e->previousorigin, d);

		state->origin[0] = e->previousorigin[0] + d[0] * blend;
		state->origin[1] = e->previousorigin[1] + d[1] * blend;
		state->origin[2] = e->previousorigin[2] + d[2] * blend;

		// rotation
		VectorSubtract (e->currentangles, e->previousangles, d);

		for (i = 0; i < 3; i++)
		{
			if (d[i] > 180)  d[i] -= 360;
			if (d[i] < -180) d[i] += 360;
		}

		state->angles[0] = e->previousangles[0] + d[0] * blend;
		state->angles[1] = e->previousangles[1] + d[1] * blend;
		state->angles[2] = e->previousangles[2] + d[2] * blend;
	}
	else
	{
		// don't lerp
		VectorCopy (e->origin, state->origin);
		VectorCopy (e->angles, state->angles);
	}
}


void GLARB_PadAliasTMU (gltexture_t *tx, int local)
{
	if (tx)
	{
		float scale[4] = {1, 1, 1, 1};

        if (tx->flags & TEXPREF_PAD)
        {
            scale[0] = (float) tx->source_width / (float) tx->width;
            scale[1] = (float) tx->source_height / (float) tx->height;
        }

		qglProgramLocalParameter4fvARB (GL_VERTEX_PROGRAM_ARB, local, scale);
	}
}


void GLARB_SetAliasTextures (entity_t *e, aliashdr_t *hdr, aliasstate_t *state)
{
	int anim = (int) (cl.time * 10) & 3;

	// dead bodies retain correct colour; swiped from QuakeWorld; this was a lot easier than i thought it would be...
	if (e->scoreboard && playertextures[e->scoreboard->playertexture])
		state->tx = playertextures[e->scoreboard->playertexture];
	else state->tx = hdr->gltextures[e->skinnum][anim];

	// ensure we have a valid texture
	if (!state->tx) state->tx = notexture;

	if (!gl_fullbrights.value)
		state->fb = NULL;
	else state->fb = hdr->fbtextures[e->skinnum][anim];

	GLARB_PadAliasTMU (state->tx, 2);
	GL_BindTexture (GL_TEXTURE0, state->tx);

	if (state->fb)
	{
		GLARB_PadAliasTMU (state->fb, 3);
		GL_BindTexture (GL_TEXTURE1, state->fb);

		if (Fog_GetDensity () > 0)
			GL_SetFragmentProgram (arb_aliasfb_fp_fog);
		else GL_SetFragmentProgram (arb_aliasfb_fp);
	}
	else
	{
		if (Fog_GetDensity () > 0)
			GL_SetFragmentProgram (arb_alias_fp_fog);
		else GL_SetFragmentProgram (arb_alias_fp);
	}
}


void *GLARB_GetXYZOffset (entity_t *e, aliashdr_t *hdr, int pose)
{
	return (void *) (hdr->vboxyzofs + hdr->numverts * pose * sizeof (meshxyz_t));
}


void *GLARB_GetNormalOffset (entity_t *e, aliashdr_t *hdr, int pose)
{
	return (void *) (hdr->vboxyzofs + (hdr->numverts * pose * sizeof (meshxyz_t)) + 4);
}


void GLMesh_Transform (entity_t *e, aliashdr_t *hdr, float *origin, float *angles)
{
	GL_TranslateMatrix (&e->matrix, origin[0], origin[1], origin[2]);

	if (angles[1]) GL_RotateMatrix (&e->matrix, angles[1], 0, 0, 1);
	if (angles[0]) GL_RotateMatrix (&e->matrix, -angles[0], 0, 1, 0);
	if (angles[2]) GL_RotateMatrix (&e->matrix, angles[2], 1, 0, 0);

	GL_TranslateMatrix (&e->matrix, hdr->scale_origin[0], hdr->scale_origin[1], hdr->scale_origin[2]);
	GL_ScaleMatrix (&e->matrix, hdr->scale[0], hdr->scale[1], hdr->scale[2]);

	// DP_ENT_SCALE
	if (e->scale != 16) GL_ScaleMatrix (&e->matrix, (float) e->scale / 16.0f, (float) e->scale / 16.0f, (float) e->scale / 16.0f);
}


void GLARB_DrawAliasFrame (entity_t *e, aliashdr_t *hdr, aliasstate_t *state, float *blendweights, qboolean viewmodel)
{
	int i;

	unsigned short *vertindexes = (unsigned short *) ((byte *) hdr + hdr->vertindexes);

	qboolean *lerptable = (qboolean *) ((byte *) hdr + hdr->lerptable);

	if (e->efrag && !(e->model->flags & EF_ROTATE))
	{
		glLoadMatrixf (r_world_matrix.m16);
		glMultMatrixf (e->matrix.m16);
	}
	else
	{
		// go back to the world matrix
		GL_LoadMatrix (&e->matrix, &r_world_matrix);
		GLMesh_Transform (e, hdr, state->origin, state->angles);
		glLoadMatrixf (e->matrix.m16);
	}

	if (viewmodel)
	{
		GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2 | VAA_3 | VAA_4 | VAA_5);
		GL_VertexArrayPointer (0, 2, 1, GL_FLOAT, GL_FALSE, 0, blendweights);
		GL_SetVertexProgram (arb_alias_vm_vp);

		for (i = 0; i < hdr->numverts; i++)
			blendweights[i] = lerptable[vertindexes[i]] ? state->blend : 1;

		// hack the depth range to prevent view model from poking into walls
		qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 1, 1, 1, 0.3, 1);
	}
	else
	{
		GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_3 | VAA_4 | VAA_5);
		GL_SetVertexProgram (arb_alias_vp);

		qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 1, state->blend, 0, 0, 0);
	}

	GLARB_SetAliasTextures (e, hdr, state);

	qglProgramLocalParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, state->shadelight);
	qglProgramLocalParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 4, state->shadevector);

	GL_VertexArrayPointer (r_meshvbo, 0, 4, GL_UNSIGNED_BYTE, GL_FALSE, sizeof (meshxyz_t), GLARB_GetXYZOffset (e, hdr, state->currpose));
	GL_VertexArrayPointer (r_meshvbo, 3, 4, GL_UNSIGNED_BYTE, GL_FALSE, sizeof (meshxyz_t), GLARB_GetXYZOffset (e, hdr, state->prevpose));
	GL_VertexArrayPointer (r_meshvbo, 1, 2, GL_FLOAT, GL_FALSE, 0, (void *) hdr->vbostofs);

	GL_VertexArrayPointer (r_meshvbo, 4, 3, GL_FLOAT, GL_FALSE, sizeof (meshxyz_t), GLARB_GetNormalOffset (e, hdr, state->currpose));
	GL_VertexArrayPointer (r_meshvbo, 5, 3, GL_FLOAT, GL_FALSE, sizeof (meshxyz_t), GLARB_GetNormalOffset (e, hdr, state->prevpose));

	GL_DrawElements (r_meshindexesvbo, GL_TRIANGLES, 0, hdr->numverts, hdr->numindexes, GL_UNSIGNED_SHORT, (void *) hdr->vboindexofs);
	rs_aliaspolys += hdr->numtris;	// consistent with ID Quake
}


void GLARB_DrawAliasBatches (entity_t **ents, int numents)
{
	int i;
	float *blendweights;

	if (!numents) return;
	if (!ents) return;

	if (gl_smoothmodels.value) glShadeModel (GL_SMOOTH);
	if (gl_affinemodels.value) glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	// take a region after the entities for temp use by the shader data and set up it's vertex pointer now (which is common)
	// so that we won't need to switch buffers while drawing the batch
	blendweights = (float *) (scratchbuf + (numents * sizeof (entity_t *)));

	for (i = 0; i < numents; i++)
	{
		entity_t *ent = ents[i];
		aliasstate_t *state = &ent->aliasstate;
		aliashdr_t *hdr = Mod_Extradata (ent->model);

		// this is a little misnamed as what it actually does is call R_CullBox
		GLARB_SetupAliasModel (ent);

		// fixme - where on earth DOES R_CullBox get called from in the old code ?????
		// we need a separate test for culling here as r_shadows mode doesn't cull
		if (ent->visframe != r_framecount) continue;

		GLARB_SetupAliasLighting (ent, state);
		GLARB_DrawAliasFrame (ent, hdr, state, blendweights, (ent == &cl.viewent));
	}

	if (gl_affinemodels.value) glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	if (gl_smoothmodels.value) glShadeModel (GL_FLAT);

	// go back to the world matrix
	glLoadMatrixf (r_world_matrix.m16);
}


// johnfitz -- values for shadow matrix
#define SHADOW_SKEW_X -0.7 // skew along x axis. -0.7 to mimic glquake shadows
#define SHADOW_SKEW_Y 0 // skew along y axis. 0 to mimic glquake shadows
#define SHADOW_VSCALE 0 // 0=completely flat
#define SHADOW_HEIGHT 0.1 // how far above the floor to render the shadow
// johnfitz

// prevent barfing on gcc
float r_shadowmatrixdata[] =
{
	1,				0,				0,				0,
	0,				1,				0,				0,
	SHADOW_SKEW_X,	SHADOW_SKEW_Y,	SHADOW_VSCALE,	0,
	0,				0,				SHADOW_HEIGHT,	1
};


glmatrix *r_shadowmatrix = (glmatrix *) r_shadowmatrixdata;


void GLARB_DrawAliasShadow (entity_t *e, aliashdr_t *hdr, aliasstate_t *state, float *blendweights, qboolean viewmodel)
{
	int i;

	unsigned short *vertindexes = (unsigned short *) ((byte *) hdr + hdr->vertindexes);

	qboolean *lerptable = (qboolean *) ((byte *) hdr + hdr->lerptable);

	glmatrix eshadow;

	// go back to the world matrix
	GL_LoadMatrix (&eshadow, &r_world_matrix);

	GL_TranslateMatrix (&eshadow, state->origin[0], state->origin[1], state->origin[2]);

	GL_TranslateMatrix (&eshadow, 0, 0, -(state->origin[2] - state->lightspot[2]));
	GL_MultiplyMatrix (&eshadow, r_shadowmatrix, &eshadow);
	GL_TranslateMatrix (&eshadow, 0, 0, (state->origin[2] - state->lightspot[2]));

	// only rotate in the horizontal plane: http://www.quake-1.com/docs/quakesrc.org/100.html
	GL_RotateMatrix (&eshadow, state->angles[1], 0, 0, 1);

	GL_TranslateMatrix (&eshadow, hdr->scale_origin[0], hdr->scale_origin[1], hdr->scale_origin[2]);
	GL_ScaleMatrix (&eshadow, hdr->scale[0], hdr->scale[1], hdr->scale[2]);

	// DP_ENT_SCALE
	if (e->scale != 16)
		GL_ScaleMatrix (&eshadow, (float) e->scale / 16.0f, (float) e->scale / 16.0f, (float) e->scale / 16.0f);

	glLoadMatrixf (eshadow.m16);

	if (viewmodel)
	{
		GL_SetVertexProgram (arb_shadow_vm_vp);
		GL_EnableVertexArrays (VAA_0 | VAA_2 | VAA_3);
		GL_VertexArrayPointer (0, 2, 1, GL_FLOAT, GL_FALSE, 0, blendweights);

		// set up lighting normals and blendweights
		for (i = 0; i < hdr->numverts; i++)
			blendweights[i] = lerptable[vertindexes[i]] ? state->blend : 1;

		// hack the depth range to prevent view model from poking into walls
		qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 1, 1, 1, 0.3, 1);
	}
	else
	{
		GL_SetVertexProgram (arb_shadow_vp);
		GL_EnableVertexArrays (VAA_0 | VAA_3);

		qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 1, state->blend, 0, 0, 0);
	}

	if (Fog_GetDensity () > 0)
		GL_SetFragmentProgram (arb_shadow_fp_fog);
	else GL_SetFragmentProgram (arb_shadow_fp);

	qglProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 0, 0, 0, 0, (r_shadows.value < 1 ? r_shadows.value : 1));

	GL_VertexArrayPointer (r_meshvbo, 0, 4, GL_UNSIGNED_BYTE, GL_FALSE, sizeof (meshxyz_t), GLARB_GetXYZOffset (e, hdr, state->currpose));
	GL_VertexArrayPointer (r_meshvbo, 3, 4, GL_UNSIGNED_BYTE, GL_FALSE, sizeof (meshxyz_t), GLARB_GetXYZOffset (e, hdr, state->prevpose));

	GL_DrawElements (r_meshindexesvbo, GL_TRIANGLES, 0, hdr->numverts, hdr->numindexes, GL_UNSIGNED_SHORT, (void *) hdr->vboindexofs);
	rs_aliaspolys += hdr->numtris;	// consistent with ID Quake
}


void GLARB_DrawShadowBatches  (entity_t **ents, int numents)
{
	int i;
	extern int gl_stencilbits;
	float *blendweights;

	if (!numents) return;
	if (!ents) return;
	if (r_shadows.value <= 0) return;

	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);

	// take a region after the entities for temp use by the shader data and set up it's vertex pointer now (which is common)
	// so that we won't need to switch buffers while drawing the batch
	blendweights = (float *) (scratchbuf + (numents * sizeof (entity_t *)));

	if (gl_stencilbits)
	{
		glEnable (GL_STENCIL_TEST);
		glStencilFunc (GL_EQUAL, 1, 2);
		glStencilOp (GL_KEEP, GL_KEEP, GL_INCR);
	}

	for (i = 0; i < numents; i++)
	{
		entity_t *ent = ents[i];
		aliasstate_t *state = &ent->aliasstate;
		aliashdr_t *hdr = Mod_Extradata (ent->model);

		if (ent->visframe != r_framecount) continue;
		if (ent->model->flags & MOD_NOSHADOW) continue;

		GLARB_DrawAliasShadow (ent, hdr, state, blendweights, (ent == &cl.viewent));
	}

	if (gl_stencilbits)
		glDisable (GL_STENCIL_TEST);

	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	// go back to the world matrix
	glLoadMatrixf (r_world_matrix.m16);
}


void GLMesh_LoadVertexBuffers (void)
{
	int j;
	model_t *m;
	int totalindexes = 0;
	int totalvbosize = 0;

	// pass 1 - count the sizes we need
	for (j = 1; j < MAX_MODELS; j++)
	{
		aliashdr_t *hdr;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_alias) continue;

		hdr = Mod_Extradata (m);

		hdr->vboindexofs = (totalindexes * sizeof (unsigned short));
		totalindexes += hdr->numindexes;

		hdr->vboxyzofs = totalvbosize;
		totalvbosize += (hdr->nummeshframes * hdr->numverts * sizeof (meshxyz_t));

		hdr->vbostofs = totalvbosize;
		totalvbosize += (hdr->numverts * sizeof (meshst_t));
	}

	if (!totalindexes) return;
	if (!totalvbosize) return;

	// pass 2 - create the buffers
	r_meshindexesvbo = GL_GetBuffer ();
	qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, r_meshindexesvbo);
	qglBufferData (GL_ELEMENT_ARRAY_BUFFER, totalindexes * sizeof (unsigned short), NULL, GL_STATIC_DRAW);

	r_meshvbo = GL_GetBuffer ();
	qglBindBuffer (GL_ARRAY_BUFFER, r_meshvbo);
	qglBufferData (GL_ARRAY_BUFFER, totalvbosize, NULL, GL_STATIC_DRAW);

	// pass 3 - fill in the buffers
	for (j = 1; j < MAX_MODELS; j++)
	{
		int f;
		aliashdr_t *hdr;
		aliasmesh_t *desc;
		meshst_t *st;

		if (!(m = cl.model_precache[j])) break;
		if (m->type != mod_alias) continue;

		hdr = Mod_Extradata (m);
		desc = (aliasmesh_t *) ((byte *) hdr + hdr->meshdesc);

		qglBufferSubData (GL_ELEMENT_ARRAY_BUFFER,
			hdr->vboindexofs,
			hdr->numindexes * sizeof (unsigned short),
			((byte *) hdr + hdr->indexes));

		for (f = 0; f < hdr->nummeshframes; f++)
		{
			int v;
			meshxyz_t *xyz = (meshxyz_t *) scratchbuf;
			trivertx_t *tv = (trivertx_t *) ((byte *) hdr + hdr->vertexes + hdr->framevertexsize * f);

			for (v = 0; v < hdr->numverts; v++)
			{
				xyz[v].xyz[0] = tv[desc[v].vertindex].v[0];
				xyz[v].xyz[1] = tv[desc[v].vertindex].v[1];
				xyz[v].xyz[2] = tv[desc[v].vertindex].v[2];
				xyz[v].xyz[3] = 1;	// need w 1 for 4 byte vertex compression

				xyz[v].normal[0] = r_avertexnormals[tv[desc[v].vertindex].lightnormalindex][0];
				xyz[v].normal[1] = r_avertexnormals[tv[desc[v].vertindex].lightnormalindex][1];
				xyz[v].normal[2] = r_avertexnormals[tv[desc[v].vertindex].lightnormalindex][2];
			}

			qglBufferSubData (GL_ARRAY_BUFFER,
				hdr->vboxyzofs + (f * hdr->numverts * sizeof (meshxyz_t)),
				hdr->numverts * sizeof (meshxyz_t),
				xyz);
		}

		st = (meshst_t *) scratchbuf;

		for (f = 0; f < hdr->numverts; f++)
		{
			st[f].st[0] = ((float) desc[f].st[0] + 0.5f) / (float) hdr->skinwidth;
			st[f].st[1] = ((float) desc[f].st[1] + 0.5f) / (float) hdr->skinheight;
		}

		qglBufferSubData (GL_ARRAY_BUFFER,
			hdr->vbostofs,
			hdr->numverts * sizeof (meshst_t),
			st);
	}

	qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, 0);
	qglBindBuffer (GL_ARRAY_BUFFER, 0);
}





