/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"

// program.local[0] = up vector
// program.local[1] = right vector
// program.local[2] = -(r_origin * vpn) vector
// program.local[3] = vpn vector
// program.local[4] = scale vector
// program.local[5] ... max = corona positions and colours
#define corona_vp \
	"!!ARBvp1.0\n" \
	"TEMP pscale, ppos;\n" \
	"TEMP localup, localright;\n" \
	"MUL localup, program.local[0], vertex.attrib[3].x;\n" \
	"MUL localright, program.local[1], vertex.attrib[3].x;\n" \
	"MAD pscale, vertex.attrib[0], program.local[3], program.local[2];\n" \
	"DPH pscale.w, pscale, program.local[4];\n" \
	"MUL pscale, pscale.w, vertex.attrib[2];\n" \
	"MAD ppos, localup, pscale.x, vertex.attrib[0];\n" \
	"MAD ppos, localright, pscale.y, ppos;\n" \
	TRANSFORMVERTEX ("result.position", "ppos") \
	"MOV result.texcoord[0], vertex.attrib[2];\n" \
	"MOV result.color, vertex.attrib[1];\n" \
	"END\n"

#define corona_fp \
	"!!ARBfp1.0\n" \
	"TEMP col0;\n" \
	"MOV col0, fragment.color;\n" \
	"DP3 col0.w, fragment.texcoord[0], fragment.texcoord[0];\n" \
	"SUB col0.w, 1.0, col0.w;\n" \
	"MOV result.color, col0;\n" \
	"END\n"

GLuint arb_corona_vp = 0;
GLuint arb_corona_fp = 0;


void Corona_CreateShaders (void)
{
	arb_corona_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, corona_vp, false);
	arb_corona_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, corona_fp, false);
}


extern cvar_t r_flatlightstyles; // johnfitz

int r_numcoronaverts = 0;

typedef struct r_corona_s
{
	float xyz[3];
	float rgba[4];
	float st[2];
	float r;
} r_corona_t;

r_corona_t *r_coronaverts = NULL;


void R_RenderCoronaVerts (void)
{
	if (r_numcoronaverts)
		GL_DrawArrays (GL_QUADS, 0, r_numcoronaverts);

	r_coronaverts = (r_corona_t *) scratchbuf;
	r_numcoronaverts = 0;
}


qboolean firstcorona = true;

void R_CoronaVertex (r_corona_t *vert, float *org, float *rgba, float s, float t, float rad)
{
	vert->xyz[0] = org[0];
	vert->xyz[1] = org[1];
	vert->xyz[2] = org[2];

	vert->rgba[0] = rgba[0];
	vert->rgba[1] = rgba[1];
	vert->rgba[2] = rgba[2];
	vert->rgba[3] = rgba[3];

	vert->st[0] = s;
	vert->st[1] = t;
	vert->r = rad;
}


void R_RenderCorona (dlight_t *light)
{
	vec3_t	v;
	float	rad;
	float	color[4];

	if (r_numcoronaverts > 8000)
		R_RenderCoronaVerts ();

	// reduce the corona size so that it coexists more peacefully with proper light
	rad = light->radius * 0.15;
	VectorSubtract (light->origin, r_origin, v);

	if (Length (v) < rad)
	{
		// view is inside the dlight
		// don;t bother adjusting the blend here because it screws things up
		return;
	}

	color[0] = light->rgb[0] * 0.1f;
	color[1] = light->rgb[1] * 0.1f;
	color[2] = light->rgb[2] * 0.1f;

	// ati throws a hairy fit with a 3-component color array
	// we don't bother making this a broken driver option as it's quite harmless, really
	color[3] = 1;

	R_CoronaVertex (&r_coronaverts[0], light->origin, color, -1, -1, rad);
	R_CoronaVertex (&r_coronaverts[1], light->origin, color, 1, -1, rad);
	R_CoronaVertex (&r_coronaverts[2], light->origin, color, 1, 1, rad);
	R_CoronaVertex (&r_coronaverts[3], light->origin, color, -1, 1, rad);

	r_numcoronaverts += 4;
	r_coronaverts += 4;
}


void R_AddCoronaToAlphaList (dlight_t *l);

cvar_t r_show_coronas = {"r_show_coronas", "1", true};

void R_CoronasBegin (void)
{
    Fog_DisableGlobalFog ();

#define PARTICLE_SCALE	1.0f
#define HACK_SCALE_UP	0.001f

	GL_SetVertexProgram (arb_corona_vp);
	GL_SetFragmentProgram (arb_corona_fp);

	qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 0, vup[0] * PARTICLE_SCALE, vup[1] * PARTICLE_SCALE, vup[2] * PARTICLE_SCALE, 0);
	qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 1, vright[0] * PARTICLE_SCALE, vright[1] * PARTICLE_SCALE, vright[2] * PARTICLE_SCALE, 0);
	qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 2, -(r_origin[0] * vpn[0]), -(r_origin[1] * vpn[1]), -(r_origin[2] * vpn[2]), 0);
	qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 3, vpn[0], vpn[1], vpn[2], 0);

	// hack a scale up to prevent particles from disappearing
	qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 4, HACK_SCALE_UP, HACK_SCALE_UP, HACK_SCALE_UP, 1.0f);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE);

	// this doesn't render at this stage but does set up the pointers/etc
	// needs to be done before we set up our vertex arrays so that r_coronaindexes and r_coronaverts point to the right things
	R_RenderCoronaVerts ();

	GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2 | VAA_3);
	GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, sizeof (r_corona_t), r_coronaverts->xyz);
	GL_VertexArrayPointer (0, 1, 4, GL_FLOAT, GL_FALSE, sizeof (r_corona_t), r_coronaverts->rgba);
	GL_VertexArrayPointer (0, 2, 2, GL_FLOAT, GL_FALSE, sizeof (r_corona_t), r_coronaverts->st);
	GL_VertexArrayPointer (0, 3, 1, GL_FLOAT, GL_FALSE, sizeof (r_corona_t), &r_coronaverts->r);
}


void R_CoronasEnd (void)
{
	// this one just renders; there's no more need for the pointers
	R_RenderCoronaVerts ();

	// fixme - make fog an enable-only thing and enable/disable during setup
	Fog_EnableGlobalFog ();

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}


void R_DrawCoronas (void)
{
	int i;
	dlight_t *l = cl_dlights;

	if (!r_show_coronas.value) return;

	for (i = 0; i < MAX_DLIGHTS; i++, l++)
	{
		if (l->die < cl.time || !l->radius)
			continue;

		R_AddCoronaToAlphaList (l);
	}
}


/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (void)
{
	// interpolated light animations
	int			j;
	float		l;
	int			flight;
	int			clight;
	float		lerpfrac;
	float		backlerp;

	// light animations
	// 'm' is normal light, 'a' is no light, 'z' is double bright
	flight = (int) floor (cl.time * 10.0f);
	clight = (int) ceil (cl.time * 10.0f);
	lerpfrac = (cl.time * 10.0f) - flight;
	backlerp = 1.0f - lerpfrac;

	for (j = 0; j < MAX_LIGHTSTYLES; j++)
	{
		if (!cl_lightstyle[j].length)
		{
			d_lightstylevalue[j] = 2.0f;
			continue;
		}
		else if (cl_lightstyle[j].length == 1)
		{
			d_lightstylevalue[j] = ((float) (cl_lightstyle[j].map[0] - 'a') * 42.5f) / 255.0f;
			continue;
		}

		l = (float) (cl_lightstyle[j].map[flight % cl_lightstyle[j].length] - 'a') * backlerp;
		l += (float) (cl_lightstyle[j].map[clight % cl_lightstyle[j].length] - 'a') * lerpfrac;

		d_lightstylevalue[j] = (l * 42.5f) / 255.0f;
	}
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/


void R_TransformDLight (glmatrix *m, float *transformed, float *origin)
{
	if (m)
	{
		// transformed
		transformed[0] = origin[0] * m->m16[0] + origin[1] * m->m16[4] + origin[2] * m->m16[8] + m->m16[12];
		transformed[1] = origin[0] * m->m16[1] + origin[1] * m->m16[5] + origin[2] * m->m16[9] + m->m16[13];
		transformed[2] = origin[0] * m->m16[2] + origin[1] * m->m16[6] + origin[2] * m->m16[10] + m->m16[14];
	}
	else
	{
		// untransformed
		transformed[0] = origin[0];
		transformed[1] = origin[1];
		transformed[2] = origin[2];
	}
}


// bumped when we go to a new entity in the current refresh
int r_dlightframecount = 0;

void R_MarkLights (dlight_t *light, int num, mnode_t *node, msurface_t *surfaces, entity_t *ent)
{
	int i;
	float dist;
	mplane_t *splitplane;
	msurface_t *surf;

	if (node->contents < 0) return;
	// if (node->visframe != r_visframecount) return;

	splitplane = node->plane;
	dist = DotProduct (light->transformed, splitplane->normal) - splitplane->dist;

	if (dist > light->radius)
	{
		R_MarkLights (light, num, node->children[0], surfaces, ent);
		return;
	}

	if (dist < -light->radius)
	{
		R_MarkLights (light, num, node->children[1], surfaces, ent);
		return;
	}

	// mark the polygons
	surf = surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// no lights on these
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		if (surf->dlightframe != r_dlightframecount)
		{
			// first time hit
			surf->dlightbits[0] = surf->dlightbits[1] = surf->dlightbits[2] = surf->dlightbits[3] = 0;
			surf->dlightframe = r_dlightframecount;
		}

		// mark the surf and ent for this dlight
		surf->dlightbits[num >> 5] |= 1 << (num & 31);
	}

	R_MarkLights (light, num, node->children[0], surfaces, ent);
	R_MarkLights (light, num, node->children[1], surfaces, ent);
}


/*
=============
R_PushDlights

made this work with any bmodel
=============
*/
void R_PushDlights (mnode_t *headnode, entity_t *ent)
{
	int i;
	dlight_t *l;
	msurface_t *surfaces;
	glmatrix *m;

	if (!cl.worldmodel->lightdata || r_fullbright.value) return;

	if (ent)
	{
		m = &ent->local;
		surfaces = ent->model->surfaces;
	}
	else
	{
		m = NULL;
		surfaces = cl.worldmodel->surfaces;
	}

	r_dlightframecount++;

	for (i = 0, l = cl_dlights; i < MAX_DLIGHTS; i++, l++)
	{
		if (l->die < cl.time || !l->radius)
			continue;

		R_TransformDLight (m, l->transformed, l->origin);
		R_MarkLights (l, i, headnode, surfaces, ent);
	}
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;
msurface_t		*lightsurf;

void R_LightSurface (msurface_t *surf, float *color)
{
	int ds = (int) ((float) DotProduct (lightspot, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]);
	int dt = (int) ((float) DotProduct (lightspot, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]);

	ds -= surf->texturemins[0];
	dt -= surf->texturemins[1];

	if (surf->samples)
	{
		// LordHavoc: enhanced to interpolate lighting
		byte *lightmap;
		int maps, line3, dsfrac = ds & 15, dtfrac = dt & 15, r00 = 0, g00 = 0, b00 = 0, r01 = 0, g01 = 0, b01 = 0, r10 = 0, g10 = 0, b10 = 0, r11 = 0, g11 = 0, b11 = 0;
		float scale;
		line3 = ((surf->extents[0] >> 4) + 1) * 3;

		lightmap = surf->samples + ((dt >> 4) * ((surf->extents[0] >> 4) + 1) + (ds >> 4)) * 3; // LordHavoc: *3 for color

		for (maps = 0; maps < MAX_SURFACE_STYLES && surf->styles[maps] != 255; maps++)
		{
			// keep this consistent with BSP lighting
			scale = d_lightstylevalue[surf->styles[maps]] * 127.5f;

			r00 += (float) lightmap[      0] * scale; g00 += (float) lightmap[      1] * scale; b00 += (float) lightmap[2] * scale;
			r01 += (float) lightmap[      3] * scale; g01 += (float) lightmap[      4] * scale; b01 += (float) lightmap[5] * scale;
			r10 += (float) lightmap[line3+0] * scale; g10 += (float) lightmap[line3+1] * scale; b10 += (float) lightmap[line3+2] * scale;
			r11 += (float) lightmap[line3+3] * scale; g11 += (float) lightmap[line3+4] * scale; b11 += (float) lightmap[line3+5] * scale;

			lightmap += ((surf->extents[0] >> 4) + 1) * ((surf->extents[1] >> 4) + 1) * 3; // LordHavoc: *3 for colored lighting
		}

		color[0] += (float) ((int) ((((((((r11 - r10) * dsfrac) >> 4) + r10) - ((((r01 - r00) * dsfrac) >> 4) + r00)) * dtfrac) >> 4) + ((((r01 - r00) * dsfrac) >> 4) + r00)));
		color[1] += (float) ((int) ((((((((g11 - g10) * dsfrac) >> 4) + g10) - ((((g01 - g00) * dsfrac) >> 4) + g00)) * dtfrac) >> 4) + ((((g01 - g00) * dsfrac) >> 4) + g00)));
		color[2] += (float) ((int) ((((((((b11 - b10) * dsfrac) >> 4) + b10) - ((((b01 - b00) * dsfrac) >> 4) + b00)) * dtfrac) >> 4) + ((((b01 - b00) * dsfrac) >> 4) + b00)));
	}
}


/*
=============
RecursiveLightPoint -- johnfitz -- replaced entire function for lit support via lordhavoc
=============
*/
int RecursiveLightPoint (float *color, mnode_t *node, vec3_t start, vec3_t end)
{
	float		front, back, frac;
	vec3_t		mid;

loc0:;
	// check for a hit
	if (node->contents < 0) return node->contents;

	// calculate mid point
	if (node->plane->type < 3)
	{
		front = start[node->plane->type] - node->plane->dist;
		back = end[node->plane->type] - node->plane->dist;
	}
	else
	{
		front = DotProduct (start, node->plane->normal) - node->plane->dist;
		back = DotProduct (end, node->plane->normal) - node->plane->dist;
	}

	// LordHavoc: optimized recursion
	if ((back < 0) == (front < 0))
		// 		return RecursiveLightPoint (color, node->children[front < 0], start, end);
	{
		node = node->children[front < 0];
		goto loc0;
	}

	frac = front / (front - back);
	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side
	if (RecursiveLightPoint (color, node->children[front < 0], start, mid) > 0)
		return true;	// hit something
	else
	{
		int i, ds, dt;
		msurface_t *surf;

		// check for impact on this node
		VectorCopy (mid, lightspot);
		lightplane = node->plane;

		surf = cl.worldmodel->surfaces + node->firstsurface;

		for (i = 0; i < node->numsurfaces; i++, surf++)
		{
			if (surf->flags & SURF_DRAWTILED)
				continue;	// no lightmaps

			ds = (int) ((float) DotProduct (mid, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]);
			dt = (int) ((float) DotProduct (mid, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]);

			if (ds < surf->texturemins[0] || dt < surf->texturemins[1])
				continue;

			ds -= surf->texturemins[0];
			dt -= surf->texturemins[1];

			if (ds > surf->extents[0] || dt > surf->extents[1])
				continue;

			if (surf->samples)
			{
				// LordHavoc: enhanced to interpolate lighting
				byte *lightmap;
				int maps, line3, dsfrac = ds & 15, dtfrac = dt & 15, r00 = 0, g00 = 0, b00 = 0, r01 = 0, g01 = 0, b01 = 0, r10 = 0, g10 = 0, b10 = 0, r11 = 0, g11 = 0, b11 = 0;
				float scale;
				line3 = surf->smax * 3;

				lightmap = surf->samples + ((dt >> 4) * surf->smax + (ds >> 4)) * 3; // LordHavoc: *3 for color

				for (maps = 0; maps < MAX_SURFACE_STYLES && surf->styles[maps] != 255; maps++)
				{
					// keep this consistent with BSP lighting
					scale = (float) d_lightstylevalue[surf->styles[maps]] * 127.5f;

					r00 += (float) lightmap[      0] * scale; g00 += (float) lightmap[      1] * scale; b00 += (float) lightmap[2] * scale;
					r01 += (float) lightmap[      3] * scale; g01 += (float) lightmap[      4] * scale; b01 += (float) lightmap[5] * scale;
					r10 += (float) lightmap[line3+0] * scale; g10 += (float) lightmap[line3+1] * scale; b10 += (float) lightmap[line3+2] * scale;
					r11 += (float) lightmap[line3+3] * scale; g11 += (float) lightmap[line3+4] * scale; b11 += (float) lightmap[line3+5] * scale;

					lightmap += surf->smax * surf->tmax * 3; // LordHavoc: *3 for colored lighting
				}

				color[0] += (float) ((int) ((((((((r11 - r10) * dsfrac) >> 4) + r10) - ((((r01 - r00) * dsfrac) >> 4) + r00)) * dtfrac) >> 4) + ((((r01 - r00) * dsfrac) >> 4) + r00)));
				color[1] += (float) ((int) ((((((((g11 - g10) * dsfrac) >> 4) + g10) - ((((g01 - g00) * dsfrac) >> 4) + g00)) * dtfrac) >> 4) + ((((g01 - g00) * dsfrac) >> 4) + g00)));
				color[2] += (float) ((int) ((((((((b11 - b10) * dsfrac) >> 4) + b10) - ((((b01 - b00) * dsfrac) >> 4) + b00)) * dtfrac) >> 4) + ((((b01 - b00) * dsfrac) >> 4) + b00)));
			}

			lightsurf = surf;
			return 1; // success
		}

		// go down back side
		return RecursiveLightPoint (color, node->children[front >= 0], mid, end);
	}
}

/*
=============
R_LightPoint -- johnfitz -- replaced entire function for lit support via lordhavoc
=============
*/
int R_LightPoint (entity_t *e, float *lightcolor)
{
	vec3_t		end;
	vec3_t		start;

	if (!cl.worldmodel->lightdata || !e || !e->model || r_fullbright.value)
	{
		// keep this consistent with BSP lighting
		lightcolor[0] = lightcolor[1] = lightcolor[2] = 128;
		return 255;
	}

	// should this just use trueorigin for everything???
	if (e->model->type == mod_brush)
	{
		start[0] = e->trueorigin[0];
		start[1] = e->trueorigin[1];
		start[2] = e->trueorigin[2];
	}
	else
	{
		start[0] = e->origin[0];
		start[1] = e->origin[1];
		start[2] = e->origin[2];
	}

	end[0] = start[0];
	end[1] = start[1];
	end[2] = start[2] - 8192; // johnfitz -- was 2048

	lightcolor[0] = lightcolor[1] = lightcolor[2] = 0;
	lightsurf = NULL;

	if (e->efrag && e->aliasstate.lightsurf)
	{
		VectorCopy (e->aliasstate.lightspot, lightspot);
		R_LightSurface (e->aliasstate.lightsurf, lightcolor);
	}
	else if (!e->efrag)
		RecursiveLightPoint (lightcolor, cl.worldmodel->nodes, start, end);

	// take down to correct scale
	VectorScale (lightcolor, (1.0f / 255.0f), lightcolor);

	// add dlights
	// mh - dynamic lights should only be added if r_dynamic is on
	// for a bmodel we'll run it through the standard dynamic lighting code
	if (r_dynamic.value && e->model->type != mod_brush)
	{
		int i;
		float dist[3];
		float brightness, r, f;
		dlight_t *light;

		for (i = 0, light = cl_dlights; i < MAX_DLIGHTS; i++, light++)
		{
			if (light->die < cl.time || light->radius < 0) continue;

			VectorSubtract (e->origin, light->origin, dist);
			f = DotProduct (dist, dist) + LIGHTOFFSET;
			r = light->radius * light->radius;

			if (f < r)
			{
				brightness = r * 7.5f / f;

				lightcolor[0] += brightness * light->rgb[0];
				lightcolor[1] += brightness * light->rgb[1];
				lightcolor[2] += brightness * light->rgb[2];
			}
		}

#if 0
		int i;
		float radius, radius2;
		float f;
		dlight_t *light;

		for (i = 0, light = cl_dlights; i < MAX_DLIGHTS; i++, light++)
		{
			if (light->die < cl.time || light->radius < 0) continue;

			radius = light->radius * light->radius;
			radius2 = radius / 7.5f;	// to do - tune this value for consistency
			f = VectorDistance2 (e->origin, light->origin);

			if (f < radius)
			{
				f = (radius2 / (f + LIGHTOFFSET)) * r_dynamic.value;

				lightcolor[0] += f * light->rgb[0];
				lightcolor[1] += f * light->rgb[1];
				lightcolor[2] += f * light->rgb[2];
			}
		}
#endif
	}

	return ((lightcolor[0] + lightcolor[1] + lightcolor[2]) * (1.0f / 3.0f));
}


void R_PrepDLights (void)
{
	int i;
	dlight_t *light;

	for (i = 0, light = cl_dlights; i < MAX_DLIGHTS; i++, light++)
	{
		if (light->die < cl.time || light->radius < 0) continue;

		light->r_radius = light->radius * light->radius;
		light->r_radius2 = light->r_radius / 7.5f;

		light->r_rgb[0] = light->rgb[0] * r_dynamic.value;
		light->r_rgb[1] = light->rgb[1] * r_dynamic.value;
		light->r_rgb[2] = light->rgb[2] * r_dynamic.value;
	}
}

