/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
============================================================================================================================================
********************************************************************************************************************************************
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

Don't use str - pr_progs->strings, Use PR_SetString

Don't use pr_progs->strings + s, use PR_GetString

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
********************************************************************************************************************************************
============================================================================================================================================
*/

#include "quakedef.h"
#include <time.h>

// this is implemented in cl_tent so that we can keep the code for this all together
// needs RMQ_TE_EXTENSIONS in the QC extensions string and a new builtin (#666) declared
void PF_GetRMQEffectsVersion (void);

#define	STRINGTEMP_BUFFERS		16
#define	STRINGTEMP_LENGTH		2048
#define PR_MAX_TEMPSTRING STRINGTEMP_LENGTH	// 2001-10-25 Enhanced temp string handling by Maddes

cvar_t pr_showstuffcmd = {"pr_showstuffcmd", "0"};

static	char	pr_string_temp[STRINGTEMP_BUFFERS][STRINGTEMP_LENGTH];
static	byte	pr_string_tempindex = 0;

static char *PR_GetTempString (void)
{
	return pr_string_temp[(STRINGTEMP_BUFFERS-1) & ++pr_string_tempindex];
}

#define	RETURN_EDICT(e) (((int *) pr_progs->globals)[OFS_RETURN] = EDICT_TO_PROG(e))


// 2001-10-20 Extension System by LordHavoc  start
char *pr_extensions[] =
{
	// add the extension names here, syntax: "extensionname",
	// ensure that the list is null terminated
	"DP_QC_STRFTIME",
	"DP_QC_TRACEBOX",
	"DP_CON_SET",
	"DP_SV_DRAWONLYTOCLIENT",
	"DP_SV_NODRAWTOCLIENT",
	"DP_TE_PARTICLERAIN",
	"DP_TE_PARTICLESNOW",
	"DP_SV_CLIENTCAMERA",
	"FRIK_FILE",
	"DP_ENT_ALPHA",
	"DP_ENT_SCALE",
	"DP_QC_SINCOSSQRTPOW",
	"DP_QC_RANDOMVEC",
	"RMQ_TE_EXTENSIONS",
	NULL
};


qboolean extension_find (char *name)
{
	int	i;

	for (i = 0;; i++)
	{
		if (!pr_extensions[i]) break;
		if (!Q_strcasecmp (pr_extensions[i], name)) return true;
	}

	return false;
}


/*
=================
PF_extension_find

returns true if the extension is supported by the server
float extension_find(string name)
=================
*/

void PF_extension_find (void)
{
	G_FLOAT (OFS_RETURN) = extension_find (G_STRING(OFS_PARM0));
}
// 2001-10-20 Extension System by LordHavoc  end


/*
===============================================================================

						BUILT-IN FUNCTIONS

===============================================================================
*/

char	pr_varstring_temp[PR_MAX_TEMPSTRING];	// 2001-10-25 Enhanced temp string handling by Maddes

#define	G_STRING2(o) (PR_GetString (((int *) pr_progs->globals)[o]))

char *PF_VarString (int	first)
{
	int		i;

// 2001-10-25 Enhanced temp string handling by Maddes  start
	int		maxlen;
	char	*add;

	pr_varstring_temp[0] = 0;

	for (i = first; i < pr_progs->prargc; i++)
	{
		add = G_STRING2 (OFS_PARM0 + i * 3);
		maxlen = PR_MAX_TEMPSTRING - strlen (pr_varstring_temp) - 1;	// -1 is EndOfString

		if (maxlen > strlen(add))
		{
			strcat (pr_varstring_temp, add);
		}
		else
		{
			strncat (pr_varstring_temp, add, maxlen);
			pr_varstring_temp[PR_MAX_TEMPSTRING - 1] = 0;
			break;	// can stop here
		}
	}

	return pr_varstring_temp;
// 2001-10-25 Enhanced temp string handling by Maddes  end
}


/*
=================
PF_errror

This is a TERMINAL error, which will kill off the entire server.
Dumps self.

error(value)
=================
*/
void PF_error (void)
{
	char	*s;
	edict_t	*ed;

	s = PF_VarString (0);

	Con_Printf ("======SERVER ERROR in %s:\n%s\n",
		PR_GetString (pr_progs->xfunction->s_name), s);

	ed = PROG_TO_EDICT (pr_progs->global_struct->self);
	ED_Print (ed);

	Host_Error ("Program error");
}

/*
=================
PF_objerror

Dumps out self, then an error message.  The program is aborted and self is
removed, but the level can continue.

objerror(value)
=================
*/
void PF_objerror (void)
{
	char	*s;
	edict_t	*ed;

	s = PF_VarString (0);

	Con_Printf ("======OBJECT ERROR in %s:\n%s\n",
		PR_GetString (pr_progs->xfunction->s_name), s);

	ed = PROG_TO_EDICT (pr_progs->global_struct->self);
	ED_Print (ed);
	ED_Free (ed);

	// Host_Error ("Program error"); // johnfitz -- by design, this should not be fatal
}



/*
==============
PF_makevectors

Writes new values for v_forward, v_up, and v_right based on angles
makevectors(vector)
==============
*/
void PF_makevectors (void)
{
	AngleVectors (G_VECTOR (OFS_PARM0), pr_progs->global_struct->v_forward, pr_progs->global_struct->v_right, pr_progs->global_struct->v_up);
}

/*
=================
PF_setorigin

This is the only valid way to move an object without using the physics of the world (setting velocity and waiting).  Directly changing origin will not set internal links correctly, so clipping would be messed up.  This should be called when an object is spawned, and then only if it is teleported.

setorigin (entity, origin)
=================
*/
void PF_setorigin (void)
{
	edict_t	*e;
	float	*org;

	e = G_EDICT (OFS_PARM0);
	org = G_VECTOR (OFS_PARM1);
	VectorCopy (org, e->v.origin);
	SV_LinkEdict (e, false);
}


void SetMinMaxSize (edict_t *e, float *min, float *max, qboolean rotate)
{
	float	*angles;
	vec3_t	rmin, rmax;
	float	bounds[2][3];
	float	xvector[2], yvector[2];
	float	a;
	vec3_t	base, transformed;
	int		i, j, k, l;

	for (i = 0; i < 3; i++)
		if (min[i] > max[i])
			PR_RunError ("backwards mins/maxs");

	rotate = false;		// FIXME: implement rotation properly again

	if (!rotate)
	{
		VectorCopy (min, rmin);
		VectorCopy (max, rmax);
	}
	else
	{
		// find min / max for rotations
		angles = e->v.angles;

		a = angles[1] / 180 * M_PI;

		xvector[0] = cos (a);
		xvector[1] = sin (a);
		yvector[0] = -sin (a);
		yvector[1] = cos (a);

		VectorCopy (min, bounds[0]);
		VectorCopy (max, bounds[1]);

		rmin[0] = rmin[1] = rmin[2] = 9999;
		rmax[0] = rmax[1] = rmax[2] = -9999;

		for (i = 0; i <= 1; i++)
		{
			base[0] = bounds[i][0];

			for (j = 0; j <= 1; j++)
			{
				base[1] = bounds[j][1];

				for (k = 0; k <= 1; k++)
				{
					base[2] = bounds[k][2];

					// transform the point
					transformed[0] = xvector[0] * base[0] + yvector[0] * base[1];
					transformed[1] = xvector[1] * base[0] + yvector[1] * base[1];
					transformed[2] = base[2];

					for (l = 0; l < 3; l++)
					{
						if (transformed[l] < rmin[l])
							rmin[l] = transformed[l];

						if (transformed[l] > rmax[l])
							rmax[l] = transformed[l];
					}
				}
			}
		}
	}

	// set derived values
	VectorCopy (rmin, e->v.mins);
	VectorCopy (rmax, e->v.maxs);
	VectorSubtract (max, min, e->v.size);

	SV_LinkEdict (e, false);
}

/*
=================
PF_setsize

the size box is rotated by the current angle

setsize (entity, minvector, maxvector)
=================
*/
void PF_setsize (void)
{
	edict_t	*e;
	float	*min, *max;

	e = G_EDICT (OFS_PARM0);
	min = G_VECTOR (OFS_PARM1);
	max = G_VECTOR (OFS_PARM2);
	SetMinMaxSize (e, min, max, false);
}


/*
=================
PF_setmodel

setmodel(entity, model)
=================
*/
void PF_setmodel (void)
{
	edict_t	*e;
	char	*m, **check;
	model_t	*mod;
	int		i;

	e = G_EDICT (OFS_PARM0);
	m = G_STRING (OFS_PARM1);

	// check to see if model was properly precached
	for (i = 0, check = sv.model_precache; *check; i++, check++)
		if (!strcmp (*check, m))
			break;

	if (!*check)
		PR_RunError ("no precache: %s\n", m);

	e->v.model = PR_SetString (*check);
	e->v.modelindex = i; // SV_ModelIndex (m);

	mod = sv.models[(int) e->v.modelindex]; // Mod_ForName (m, true);

	if (mod)
		// johnfitz -- correct physics cullboxes for bmodels
	{
		if (mod->type == mod_brush)
			SetMinMaxSize (e, mod->clipmins, mod->clipmaxs, true);
		else
			SetMinMaxSize (e, mod->mins, mod->maxs, true);
	}
	// johnfitz
	else
		SetMinMaxSize (e, vec3_origin, vec3_origin, true);
}

/*
=================
PF_bprint

broadcast print to everyone on server

bprint(value)
=================
*/
void PF_bprint (void)
{
	char		*s;

	s = PF_VarString (0);
	SV_BroadcastPrintf ("%s", s);
}

/*
=================
PF_sprint

single print to a specific client

sprint(clientent, value)
=================
*/
void PF_sprint (void)
{
	char		*s;
	client_t	*client;
	int			entnum;

	entnum = G_EDICTNUM (OFS_PARM0);
	s = PF_VarString (1);

	if (entnum < 1 || entnum > svs.maxclients)
	{
		Con_Printf ("tried to sprint to a non-client\n");
		return;
	}

	client = &svs.clients[entnum-1];

	MSG_WriteChar (&client->message, svc_print);
	MSG_WriteString (&client->message, s);
}


/*
=================
PF_centerprint

single print to a specific client

centerprint(clientent, value)
=================
*/
void PF_centerprint (void)
{
	char		*s;
	client_t	*client;
	int			entnum;

	entnum = G_EDICTNUM (OFS_PARM0);
	s = PF_VarString (1);

	if (entnum < 1 || entnum > svs.maxclients)
	{
		Con_Printf ("tried to sprint to a non-client\n");
		return;
	}

	client = &svs.clients[entnum-1];

	MSG_WriteChar (&client->message, svc_centerprint);
	MSG_WriteString (&client->message, s);
}


/*
=================
PF_normalize

vector normalize(vector)
=================
*/
void PF_normalize (void)
{
	float	*value1;
	vec3_t	newvalue;
	float	new_temp;

	value1 = G_VECTOR (OFS_PARM0);

	new_temp = value1[0] * value1[0] + value1[1] * value1[1] + value1[2] * value1[2];
	new_temp = sqrt (new_temp);

	if (new_temp == 0)
		newvalue[0] = newvalue[1] = newvalue[2] = 0;
	else
	{
		new_temp = 1 / new_temp;
		newvalue[0] = value1[0] * new_temp;
		newvalue[1] = value1[1] * new_temp;
		newvalue[2] = value1[2] * new_temp;
	}

	VectorCopy (newvalue, G_VECTOR (OFS_RETURN));
}

/*
=================
PF_vlen

scalar vlen(vector)
=================
*/
void PF_vlen (void)
{
	float	*value1;
	float	new_temp;

	value1 = G_VECTOR (OFS_PARM0);

	new_temp = value1[0] * value1[0] + value1[1] * value1[1] + value1[2] * value1[2];
	new_temp = sqrt (new_temp);

	G_FLOAT (OFS_RETURN) = new_temp;
}


void PF_sin (void)
{
	float num, val;

	num = G_FLOAT (OFS_PARM0);

	// would a kind doctor please tell MSVC that this is C, not C++, do there's no need to throw errors about the C++ overloaded versions
	val = (float) sin ((float) num);

	G_FLOAT (OFS_RETURN) = val;
}


void PF_cos (void)
{
	float num, val;

	num = G_FLOAT (OFS_PARM0);

	// would a kind doctor please tell MSVC that this is C, not C++, do there's no need to throw errors about the C++ overloaded versions
	val = (float) cos ((float) num);

	G_FLOAT (OFS_RETURN) = val;
}


void PF_sqrt (void)
{
	float num, val;

	num = G_FLOAT (OFS_PARM0);

	// would a kind doctor please tell MSVC that this is C, not C++, do there's no need to throw errors about the C++ overloaded versions
	val = (float) sqrt ((float) num);

	G_FLOAT (OFS_RETURN) = val;
}


void PF_fpow (void)
{
	float base, expo, val;

	base = G_FLOAT (OFS_PARM0);
	expo = G_FLOAT (OFS_PARM1);

	// would a kind doctor please tell MSVC that this is C, not C++, do there's no need to throw errors about the C++ overloaded versions
	val = (float) pow ((float) base, (float) expo);

	G_FLOAT (OFS_RETURN) = val;
}


void PF_randomvec (void)
{
	vec3_t		temp;

	// // WTF ??
	// i'll second that - but it maintains invariance with DP behaviour so oh well.
	do
	{
		temp[0] = (rand () & 32767) * (2.0 / 32767.0) - 1.0;
		temp[1] = (rand () & 32767) * (2.0 / 32767.0) - 1.0;
		temp[2] = (rand () & 32767) * (2.0 / 32767.0) - 1.0;
	} while (DotProduct (temp, temp) >= 1);

	VectorCopy (temp, G_VECTOR (OFS_RETURN));
}


/*
=================
PF_vectoyaw

float vectoyaw(vector)
=================
*/
void PF_vectoyaw (void)
{
	float	*value1;
	float	yaw;

	value1 = G_VECTOR (OFS_PARM0);

	if (value1[1] == 0 && value1[0] == 0)
		yaw = 0;
	else
	{
		yaw = (int) (atan2 (value1[1], value1[0]) * 180 / M_PI);

		if (yaw < 0)
			yaw += 360;
	}

	G_FLOAT (OFS_RETURN) = yaw;
}


/*
=================
PF_vectoangles

vector vectoangles(vector)
=================
*/
void PF_vectoangles (void)
{
	float	*value1;
	float	forward;
	float	yaw, pitch;

	value1 = G_VECTOR (OFS_PARM0);

	if (value1[1] == 0 && value1[0] == 0)
	{
		yaw = 0;

		if (value1[2] > 0)
			pitch = 90;
		else
			pitch = 270;
	}
	else
	{
		yaw = (int) (atan2 (value1[1], value1[0]) * 180 / M_PI);

		if (yaw < 0)
			yaw += 360;

		forward = sqrt (value1[0] * value1[0] + value1[1] * value1[1]);
		pitch = (int) (atan2 (value1[2], forward) * 180 / M_PI);

		if (pitch < 0)
			pitch += 360;
	}

	G_FLOAT (OFS_RETURN + 0) = pitch;
	G_FLOAT (OFS_RETURN + 1) = yaw;
	G_FLOAT (OFS_RETURN + 2) = 0;
}

/*
=================
PF_Random

Returns a number from 0<= num < 1

random ()
=================
*/
void PF_random (void)
{
	float		num;

	num = (rand () & 0x7fff) / ((float) 0x7fff);

	G_FLOAT (OFS_RETURN) = num;
}

/*
=================
PF_particle

particle(origin, color, count)
=================
*/
void PF_particle (void)
{
	float		*org, *dir;
	float		color;
	float		count;

	org = G_VECTOR (OFS_PARM0);
	dir = G_VECTOR (OFS_PARM1);
	color = G_FLOAT (OFS_PARM2);
	count = G_FLOAT (OFS_PARM3);
	SV_StartParticle (org, dir, color, count);
}


/*
=================
PF_ambientsound

=================
*/
void PF_ambientsound (void)
{
	char		**check;
	char		*samp;
	float		*pos;
	float 		vol, attenuation;
	int			i, soundnum;
	int			large = false; // johnfitz -- PROTOCOL_FITZQUAKE

	pos = G_VECTOR (OFS_PARM0);
	samp = G_STRING (OFS_PARM1);
	vol = G_FLOAT (OFS_PARM2);
	attenuation = G_FLOAT (OFS_PARM3);

	// check to see if samp was properly precached
	for (soundnum = 0, check = sv.sound_precache; *check; check++, soundnum++)
		if (!strcmp (*check, samp))
			break;

	if (!*check)
	{
		Con_Printf ("no precache: %s\n", samp);
		return;
	}

	// johnfitz -- PROTOCOL_FITZQUAKE
	if (soundnum > 255)
	{
		if (sv.protocol == PROTOCOL_NETQUAKE)
			return; // don't send any info protocol can't support
		else large = true;
	}

	// johnfitz

	// add an svc_spawnambient command to the level signon packet

	// johnfitz -- PROTOCOL_FITZQUAKE
	if (large)
		MSG_WriteByte (&sv.signon, svc_spawnstaticsound2);
	else
		MSG_WriteByte (&sv.signon, svc_spawnstaticsound);

	// johnfitz

	for (i = 0; i < 3; i++)
		MSG_WriteCoord (&sv.signon, pos[i], sv.protocol, sv.protocolflags);	//

	// johnfitz -- PROTOCOL_FITZQUAKE
	if (large)
		MSG_WriteShort (&sv.signon, soundnum);
	else
		MSG_WriteByte (&sv.signon, soundnum);

	// johnfitz

	MSG_WriteByte (&sv.signon, vol * 255);
	MSG_WriteByte (&sv.signon, attenuation * 64);

}

/*
=================
PF_sound

Each entity can have eight independant sound sources, like voice,
weapon, feet, etc.

Channel 0 is an auto-allocate channel, the others override anything
allready running on that entity/channel pair.

An attenuation of 0 will play full volume everywhere in the level.
Larger attenuations will drop off.

=================
*/
void PF_sound (void)
{
	char		*sample;
	int			channel;
	edict_t		*entity;
	int 		volume;
	float attenuation;

	entity = G_EDICT (OFS_PARM0);
	channel = G_FLOAT (OFS_PARM1);
	sample = G_STRING (OFS_PARM2);
	volume = G_FLOAT (OFS_PARM3) * 255;
	attenuation = G_FLOAT (OFS_PARM4);

	if (volume < 0 || volume > 255)
		Sys_Error ("SV_StartSound: volume = %i", volume);

	if (attenuation < 0 || attenuation > 4)
		Sys_Error ("SV_StartSound: attenuation = %f", attenuation);

	if (channel < 0 || channel > 7)
		Sys_Error ("SV_StartSound: channel = %i", channel);

	SV_StartSound (entity, channel, sample, volume, attenuation);
}

/*
=================
PF_break

break ()
=================
*/
void PF_break (void)
{
	Con_Printf ("break statement\n");
	*(int *) -4 = 0;	// dump to debugger
	// 	PR_RunError ("break statement");
}

/*
=================
PF_traceline

Used for use tracing and shot targeting
Traces are blocked by bbox and exact bsp entityes, and also slide box entities
if the tryents flag is set.

traceline (vector1, vector2, tryents)
=================
*/
void PF_traceline (void)
{
	float	*v1, *v2;
	trace_t	trace;
	int		nomonsters;
	edict_t	*ent;

	v1 = G_VECTOR (OFS_PARM0);
	v2 = G_VECTOR (OFS_PARM1);
	nomonsters = G_FLOAT (OFS_PARM2);
	ent = G_EDICT (OFS_PARM3);

	// http://trac.assembla.com/RemakeQuake/ticket/120#comment:910
	if (v1[0] != v1[0]) v1[0] = 0;
	if (v1[1] != v1[1]) v1[1] = 0;
	if (v1[2] != v1[2]) v1[2] = 0;

	if (v2[0] != v2[0]) v2[0] = 0;
	if (v2[1] != v2[1]) v2[1] = 0;
	if (v2[2] != v2[2]) v2[2] = 0;

	trace = SV_Move (v1, vec3_origin, vec3_origin, v2, nomonsters, ent);

	pr_progs->global_struct->trace_allsolid = trace.allsolid;
	pr_progs->global_struct->trace_startsolid = trace.startsolid;
	pr_progs->global_struct->trace_fraction = trace.fraction;
	pr_progs->global_struct->trace_inwater = trace.inwater;
	pr_progs->global_struct->trace_inopen = trace.inopen;
	VectorCopy (trace.endpos, pr_progs->global_struct->trace_endpos);
	VectorCopy (trace.plane.normal, pr_progs->global_struct->trace_plane_normal);
	pr_progs->global_struct->trace_plane_dist =  trace.plane.dist;

	if (trace.ent)
		pr_progs->global_struct->trace_ent = EDICT_TO_PROG (trace.ent);
	else pr_progs->global_struct->trace_ent = EDICT_TO_PROG (pr_progs->edicts[0]);
}

/*
=================
PF_checkpos

Returns true if the given entity can move to the given position from it's
current position by walking or rolling.
FIXME: make work...
scalar checkpos (entity, vector)
=================
*/
void PF_checkpos (void)
{
}

// ============================================================================

byte	checkpvs[MAX_MAP_LEAFS/8];

int PF_newcheckclient (int check)
{
	int		i;
	byte	*pvs;
	edict_t	*ent;
	mleaf_t	*leaf;
	vec3_t	org;

	// cycle to the next one

	if (check < 1)
		check = 1;

	if (check > svs.maxclients)
		check = svs.maxclients;

	if (check == svs.maxclients)
		i = 1;
	else
		i = check + 1;

	for (; ; i++)
	{
		if (i == svs.maxclients + 1)
			i = 1;

		ent = GetEdictForNumber (i);

		if (i == check)
			break;	// didn't find anything else

		if (ent->free)
			continue;

		if (ent->v.health <= 0)
			continue;

		if ((int) ent->v.flags & FL_NOTARGET)
			continue;

		// anything that is a client, or has a client as an enemy
		break;
	}

	// get the PVS for the entity
	VectorAdd (ent->v.origin, ent->v.view_ofs, org);
	leaf = Mod_PointInLeaf (org, sv.worldmodel);
	pvs = Mod_LeafPVS (leaf, sv.worldmodel);
	memcpy (checkpvs, pvs, (sv.worldmodel->numleafs + 7) >> 3);

	return i;
}

/*
=================
PF_checkclient

Returns a client (or object that has a client enemy) that would be a
valid target.

If there are more than one valid options, they are cycled each frame

If (self.origin + self.viewofs) is not in the PVS of the current target,
it is not returned at all.

name checkclient ()
=================
*/
#define	MAX_CHECK	16
int c_invis, c_notvis;

void PF_checkclient (void)
{
	edict_t	*ent, *self;
	mleaf_t	*leaf;
	int		l;
	vec3_t	view;

	// find a new check if on a new frame
	if (pr_progs->time - sv.lastchecktime >= 0.1)
	{
		sv.lastcheck = PF_newcheckclient (sv.lastcheck);
		sv.lastchecktime = pr_progs->time;
	}

	// return check if it might be visible
	ent = GetEdictForNumber (sv.lastcheck);

	if (ent->free || ent->v.health <= 0)
	{
		RETURN_EDICT (pr_progs->edicts[0]);
		return;
	}

	// if current entity can't possibly see the check entity, return 0
	self = PROG_TO_EDICT (pr_progs->global_struct->self);
	VectorAdd (self->v.origin, self->v.view_ofs, view);
	leaf = Mod_PointInLeaf (view, sv.worldmodel);
	l = (leaf - sv.worldmodel->leafs) - 1;

	if ((l < 0) || !(checkpvs[l>>3] & (1 << (l & 7))))
	{
		c_notvis++;
		RETURN_EDICT (pr_progs->edicts[0]);
		return;
	}

	// might be able to see it
	c_invis++;
	RETURN_EDICT (ent);
}

// ============================================================================


/*
=================
PF_stuffcmd

Sends text over to the client's execution buffer

stuffcmd (clientent, value)
=================
*/
void PF_stuffcmd (void)
{
	int		entnum;
	char	*str;
	client_t	*old;

	entnum = G_EDICTNUM (OFS_PARM0);

	if (entnum < 1 || entnum > svs.maxclients)
		PR_RunError ("Parm 0 not a client");

	str = G_STRING (OFS_PARM1);

	if (pr_showstuffcmd.value)
		Con_Printf ("stuffcmd: \"%s\"\n", str);

#ifdef RMQMSVC
#ifdef _DEBUG
	// hack to prevent RMQ from autosaving in MSVC debug builds during testing when we'll be
	// constantly loading and reloading lots of maps and need to clear down the save dir every day or so
	if (!strncmp (str, "save ", 5)) return;
#endif
#endif

	old = host_client;
	host_client = &svs.clients[entnum-1];
	Host_ClientCommands ("%s", str);
	host_client = old;
}

/*
=================
PF_localcmd

Sends text over to the client's execution buffer

localcmd (string)
=================
*/
void PF_localcmd (void)
{
	char	*str;

	str = G_STRING (OFS_PARM0);
	Cbuf_AddText (str);
}

/*
=================
PF_cvar

float cvar (string)
=================
*/
void PF_cvar (void)
{
	char	*str;

	str = G_STRING (OFS_PARM0);

	G_FLOAT (OFS_RETURN) = Cvar_VariableValue (str);
}

/*
=================
PF_cvar_set

float cvar (string)
=================
*/
void PF_cvar_set (void)
{
	char	*var, *val;

	var = G_STRING (OFS_PARM0);
	val = G_STRING (OFS_PARM1);

	Cvar_Set (var, val);
}

/*
=================
PF_findradius

Returns a chain of entities that have origins within a spherical area

findradius (origin, radius)
=================
*/
void PF_findradius (void)
{
	edict_t	*ent, *chain;
	float	rad;
	float	*org;
	vec3_t	eorg;
	int		i, j;

	chain = (edict_t *) pr_progs->edicts[0];

	org = G_VECTOR (OFS_PARM0);
	rad = G_FLOAT (OFS_PARM1);

	ent = NEXT_EDICT (pr_progs->edicts[0]);

	for (i = 1; i < pr_progs->num_edicts; i++, ent = NEXT_EDICT (ent))
	{
		if (ent->free)
			continue;

		if (ent->v.solid == SOLID_NOT)
			continue;

		for (j = 0; j < 3; j++)
			eorg[j] = org[j] - (ent->v.origin[j] + (ent->v.mins[j] + ent->v.maxs[j]) * 0.5);

		if (Length (eorg) > rad)
			continue;

		ent->v.chain = EDICT_TO_PROG (chain);
		chain = ent;
	}

	RETURN_EDICT (chain);
}


/*
=========
PF_dprint
=========
*/
void PF_dprint (void)
{
	Con_DPrintf ("%s", PF_VarString (0));
}

void PF_ftos (void)
{
	float	v;
	char	*s;

	v = G_FLOAT (OFS_PARM0);
	s = PR_GetTempString ();

	if (v == (int) v)
		sprintf (s, "%d", (int) v);
	else sprintf (s, "%g", v);
		// sprintf (s, "%5.1f", v);

//	Con_Printf ("PF_ftos : %5.1f %g\n", v, v);

	G_INT (OFS_RETURN) = PR_SetString (s);
}

void PF_fabs (void)
{
	float	v;
	v = G_FLOAT (OFS_PARM0);
	G_FLOAT (OFS_RETURN) = fabs (v);
}

void PF_vtos (void)
{
	char	*s;

	s = PR_GetTempString ();
	sprintf (s, "'%5.1f %5.1f %5.1f'", G_VECTOR (OFS_PARM0)[0], G_VECTOR (OFS_PARM0)[1], G_VECTOR (OFS_PARM0)[2]);
	G_INT (OFS_RETURN) = PR_SetString (s);
}

void PF_Spawn (void)
{
	edict_t	*ed;
	ed = ED_Alloc ();
	RETURN_EDICT (ed);
}

void PF_Remove (void)
{
	edict_t	*ed;

	ed = G_EDICT (OFS_PARM0);
	ED_Free (ed);
}


// entity (entity start, .string field, string match) find = #5;
void PF_Find (void)
{
	int		e;
	int		f;
	char	*s, *t;
	edict_t	*ed;

	e = G_EDICTNUM (OFS_PARM0);
	f = G_INT (OFS_PARM1);
	s = G_STRING (OFS_PARM2);

	if (!s)
		PR_RunError ("PF_Find: bad search string");

	for (e++; e < pr_progs->num_edicts; e++)
	{
		ed = GetEdictForNumber (e);

		if (ed->free)
			continue;

		t = E_STRING (ed, f);

		if (!t)
			continue;

		if (!strcmp (t, s))
		{
			RETURN_EDICT (ed);
			return;
		}
	}

	RETURN_EDICT (pr_progs->edicts[0]);
}

void PR_CheckEmptyString (char *s)
{
	if (s[0] <= ' ')
		PR_RunError ("Bad string");
}

void PF_precache_file (void)
{
	// precache_file is only used to copy files with qcc, it does nothing
	G_INT (OFS_RETURN) = G_INT (OFS_PARM0);
}

void PF_precache_sound (void)
{
	char	*s;
	int		i;

	if (sv.state != ss_loading)
		PR_RunError ("PF_Precache_*: Precache can only be done in spawn functions");

	s = G_STRING (OFS_PARM0);
	G_INT (OFS_RETURN) = G_INT (OFS_PARM0);
	PR_CheckEmptyString (s);

	for (i = 0; i < MAX_SOUNDS; i++)
	{
		if (!sv.sound_precache[i])
		{
			sv.sound_precache[i] = s;
			return;
		}

		if (!strcmp (sv.sound_precache[i], s))
			return;
	}

	PR_RunError ("PF_precache_sound: overflow");
}

void PF_precache_model (void)
{
	char	*s;
	int		i;

	if (sv.state != ss_loading)
		PR_RunError ("PF_Precache_*: Precache can only be done in spawn functions");

	s = G_STRING (OFS_PARM0);
	G_INT (OFS_RETURN) = G_INT (OFS_PARM0);
	PR_CheckEmptyString (s);

	for (i = 0; i < MAX_MODELS; i++)
	{
		if (!sv.model_precache[i])
		{
			sv.model_precache[i] = s;
			sv.models[i] = Mod_ForName (s, true);
			return;
		}

		if (!strcmp (sv.model_precache[i], s))
			return;
	}

	PR_RunError ("PF_precache_model: overflow");
}


void PF_coredump (void)
{
	ED_PrintEdicts ();
}

void PF_traceon (void)
{
	pr_progs->trace = true;
}

void PF_traceoff (void)
{
	pr_progs->trace = false;
}

void PF_eprint (void)
{
	ED_PrintNum (G_EDICTNUM (OFS_PARM0));
}

/*
===============
PF_walkmove

float(float yaw, float dist) walkmove
===============
*/
void PF_walkmove (void)
{
	edict_t	*ent;
	float	yaw, dist;
	vec3_t	move;
	dfunction_t	*oldf;
	int 	oldself;

	ent = PROG_TO_EDICT (pr_progs->global_struct->self);
	yaw = G_FLOAT (OFS_PARM0);
	dist = G_FLOAT (OFS_PARM1);

	if (!((int) ent->v.flags & (FL_ONGROUND | FL_FLY | FL_SWIM)))
	{
		G_FLOAT (OFS_RETURN) = 0;
		return;
	}

	yaw = yaw * M_PI * 2 / 360;

	move[0] = cos (yaw) * dist;
	move[1] = sin (yaw) * dist;
	move[2] = 0;

	// save program state, because SV_movestep may call other progs
	oldf = pr_progs->xfunction;
	oldself = pr_progs->global_struct->self;

	G_FLOAT (OFS_RETURN) = SV_movestep (ent, move, true);

	// restore program state
	pr_progs->xfunction = oldf;
	pr_progs->global_struct->self = oldself;
}

/*
===============
PF_droptofloor

void () droptofloor
===============
*/
void PF_droptofloor (void)
{
	edict_t		*ent;
	vec3_t		end;
	trace_t		trace;

	ent = PROG_TO_EDICT (pr_progs->global_struct->self);

	VectorCopy (ent->v.origin, end);
	end[2] -= 256;

	trace = SV_Move (ent->v.origin, ent->v.mins, ent->v.maxs, end, false, ent);

	if (trace.fraction == 1 || trace.allsolid)
		G_FLOAT (OFS_RETURN) = 0;
	else
	{
		VectorCopy (trace.endpos, ent->v.origin);
		SV_LinkEdict (ent, false);
		ent->v.flags = (int) ent->v.flags | FL_ONGROUND;
		ent->v.groundentity = EDICT_TO_PROG (trace.ent);
		G_FLOAT (OFS_RETURN) = 1;
	}
}

/*
===============
PF_lightstyle

void(float style, string value) lightstyle
===============
*/
void PF_lightstyle (void)
{
	int		style;
	char	*val;
	client_t	*client;
	int			j;

	style = G_FLOAT (OFS_PARM0);
	val = G_STRING (OFS_PARM1);

	// change the string in sv
	sv.lightstyles[style] = val;

	// send message to all clients on this server
	if (sv.state != ss_active)
		return;

	for (j = 0, client = svs.clients; j < svs.maxclients; j++, client++)
	{
		if (client->active || client->spawned)
		{
			MSG_WriteChar (&client->message, svc_lightstyle);
			MSG_WriteChar (&client->message, style);
			MSG_WriteString (&client->message, val);
		}
	}
}

void PF_rint (void)
{
	float	f;
	f = G_FLOAT (OFS_PARM0);

	if (f > 0)
		G_FLOAT (OFS_RETURN) = (int) (f + 0.5);
	else
		G_FLOAT (OFS_RETURN) = (int) (f - 0.5);
}
void PF_floor (void)
{
	G_FLOAT (OFS_RETURN) = floor (G_FLOAT (OFS_PARM0));
}
void PF_ceil (void)
{
	G_FLOAT (OFS_RETURN) = ceil (G_FLOAT (OFS_PARM0));
}


/*
=============
PF_checkbottom
=============
*/
void PF_checkbottom (void)
{
	edict_t	*ent;

	ent = G_EDICT (OFS_PARM0);

	G_FLOAT (OFS_RETURN) = SV_CheckBottom (ent);
}

/*
=============
PF_pointcontents
=============
*/
void PF_pointcontents (void)
{
	float	*v;

	v = G_VECTOR (OFS_PARM0);

	G_FLOAT (OFS_RETURN) = SV_PointContents (v);
}

/*
=============
PF_nextent

entity nextent(entity)
=============
*/
void PF_nextent (void)
{
	int		i;
	edict_t	*ent;

	i = G_EDICTNUM (OFS_PARM0);

	while (1)
	{
		i++;

		if (i == pr_progs->num_edicts)
		{
			RETURN_EDICT (pr_progs->edicts[0]);
			return;
		}

		ent = GetEdictForNumber (i);

		if (!ent->free)
		{
			RETURN_EDICT (ent);
			return;
		}
	}
}

/*
=============
PF_aim

Pick a vector for the player to shoot along
vector aim(entity, missilespeed)
=============
*/
cvar_t	sv_aim = {"sv_aim", "0.93"};
void PF_aim (void)
{
	edict_t	*ent, *check, *bestent;
	vec3_t	start, dir, end, bestdir;
	int		i, j;
	trace_t	tr;
	float	dist, bestdist;
	float	speed;

	ent = G_EDICT (OFS_PARM0);
	speed = G_FLOAT (OFS_PARM1);

	VectorCopy (ent->v.origin, start);
	start[2] += 20;

	// try sending a trace straight
	VectorCopy (pr_progs->global_struct->v_forward, dir);
	VectorMA (start, 2048, dir, end);
	tr = SV_Move (start, vec3_origin, vec3_origin, end, false, ent);

	if (tr.ent && tr.ent->v.takedamage == DAMAGE_AIM
			&& (!teamplay.value || ent->v.team <= 0 || ent->v.team != tr.ent->v.team))
	{
		VectorCopy (pr_progs->global_struct->v_forward, G_VECTOR (OFS_RETURN));
		return;
	}


	// try all possible entities
	VectorCopy (dir, bestdir);
	bestdist = sv_aim.value;
	bestent = NULL;

	check = NEXT_EDICT (pr_progs->edicts[0]);

	for (i = 1; i < pr_progs->num_edicts; i++, check = NEXT_EDICT (check))
	{
		if (check->v.takedamage != DAMAGE_AIM)
			continue;

		if (check == ent)
			continue;

		if (teamplay.value && ent->v.team > 0 && ent->v.team == check->v.team)
			continue;	// don't aim at teammate

		for (j = 0; j < 3; j++)
			end[j] = check->v.origin[j]
					 + 0.5 * (check->v.mins[j] + check->v.maxs[j]);

		VectorSubtract (end, start, dir);
		VectorNormalize (dir);
		dist = DotProduct (dir, pr_progs->global_struct->v_forward);

		if (dist < bestdist)
			continue;	// to far to turn

		tr = SV_Move (start, vec3_origin, vec3_origin, end, false, ent);

		if (tr.ent == check)
		{
			// can shoot at this one
			bestdist = dist;
			bestent = check;
		}
	}

	if (bestent)
	{
		VectorSubtract (bestent->v.origin, ent->v.origin, dir);
		dist = DotProduct (dir, pr_progs->global_struct->v_forward);
		VectorScale (pr_progs->global_struct->v_forward, dist, end);
		end[2] = dir[2];
		VectorNormalize (end);
		VectorCopy (end, G_VECTOR (OFS_RETURN));
	}
	else
	{
		VectorCopy (bestdir, G_VECTOR (OFS_RETURN));
	}
}

/*
==============
PF_changeyaw

This was a major timewaster in progs, so it was converted to C
==============
*/
void PF_changeyaw (void)
{
	edict_t		*ent;
	float		ideal, current, move, speed;

	ent = PROG_TO_EDICT (pr_progs->global_struct->self);
	current = anglemod (ent->v.angles[1]);
	ideal = ent->v.ideal_yaw;
	speed = ent->v.yaw_speed;

	if (current == ideal)
		return;

	move = ideal - current;

	if (ideal > current)
	{
		if (move >= 180)
			move = move - 360;
	}
	else
	{
		if (move <= -180)
			move = move + 360;
	}

	if (move > 0)
	{
		if (move > speed)
			move = speed;
	}
	else
	{
		if (move < -speed)
			move = -speed;
	}

	ent->v.angles[1] = anglemod (current + move);
}

/*
===============================================================================

MESSAGE WRITING

===============================================================================
*/

#define	MSG_BROADCAST	0		// unreliable to all
#define	MSG_ONE			1		// reliable to one (msg_entity)
#define	MSG_ALL			2		// reliable to all
#define	MSG_INIT		3		// write to the init string

sizebuf_t *WriteDest (void)
{
	int		entnum;
	int		dest;
	edict_t	*ent;

	dest = G_FLOAT (OFS_PARM0);

	switch (dest)
	{
	case MSG_BROADCAST:
		return &sv.datagram;

	case MSG_ONE:
		ent = PROG_TO_EDICT (pr_progs->global_struct->msg_entity);
		entnum = GetNumberForEdict (ent);

		if (entnum < 1 || entnum > svs.maxclients)
			PR_RunError ("WriteDest: not a client");

		return &svs.clients[entnum-1].message;

	case MSG_ALL:
		return &sv.reliable_datagram;

	case MSG_INIT:
		return &sv.signon;

	default:
		PR_RunError ("WriteDest: bad destination");
		break;
	}

	return NULL;
}

void PF_WriteByte (void)
{
	MSG_WriteByte (WriteDest (), G_FLOAT (OFS_PARM1));
}

void PF_WriteChar (void)
{
	MSG_WriteChar (WriteDest (), G_FLOAT (OFS_PARM1));
}

void PF_WriteShort (void)
{
	MSG_WriteShort (WriteDest (), G_FLOAT (OFS_PARM1));
}

void PF_WriteLong (void)
{
	MSG_WriteLong (WriteDest (), G_FLOAT (OFS_PARM1));
}

void PF_WriteAngle (void)
{
	// not much we can do about this as we don't know which angle is being written...
	MSG_WriteAngle (WriteDest (), G_FLOAT (OFS_PARM1), sv.protocol, sv.protocolflags, false);
}

void PF_WriteCoord (void)
{
	MSG_WriteCoord (WriteDest (), G_FLOAT (OFS_PARM1), sv.protocol, sv.protocolflags);	//
}

void PF_WriteString (void)
{
	MSG_WriteString (WriteDest (), G_STRING (OFS_PARM1));
}


void PF_WriteEntity (void)
{
	MSG_WriteShort (WriteDest (), G_EDICTNUM (OFS_PARM1));
}

// =============================================================================

int SV_ModelIndex (char *name);

void PF_makestatic (void)
{
	edict_t		*ent;
	int			bits = 0; // johnfitz -- PROTOCOL_FITZQUAKE

	ent = G_EDICT (OFS_PARM0);

	// johnfitz -- don't send invisible static entities
	if (ent->alpha == ENTALPHA_ZERO)
	{
		ED_Free (ent);
		return;
	}

	// johnfitz

	// johnfitz -- PROTOCOL_FITZQUAKE
	if (sv.protocol == PROTOCOL_NETQUAKE)
	{
		if (SV_ModelIndex (PR_GetString (ent->v.model)) & 0xFF00 || (int) (ent->v.frame) & 0xFF00)
		{
			ED_Free (ent);
			return; // can't display the correct model & frame, so don't show it at all
		}
	}
	else
	{
		if (SV_ModelIndex (PR_GetString (ent->v.model)) & 0xFF00)
			bits |= B_LARGEMODEL;

		if ((int) (ent->v.frame) & 0xFF00)
			bits |= B_LARGEFRAME;

		if (ent->alpha != ENTALPHA_DEFAULT)
			bits |= B_ALPHA;
	}

	if (bits)
	{
		MSG_WriteByte (&sv.signon, svc_spawnstatic2);
		MSG_WriteByte (&sv.signon, bits);
	}
	else
		MSG_WriteByte (&sv.signon, svc_spawnstatic);

	if (bits & B_LARGEMODEL)
		MSG_WriteShort (&sv.signon, SV_ModelIndex (PR_GetString (ent->v.model)));
	else
		MSG_WriteByte (&sv.signon, SV_ModelIndex (PR_GetString (ent->v.model)));

	if (bits & B_LARGEFRAME)
		MSG_WriteShort (&sv.signon, ent->v.frame);
	else
		MSG_WriteByte (&sv.signon, ent->v.frame);

	// johnfitz

	MSG_WriteByte (&sv.signon, ent->v.colormap);
	MSG_WriteByte (&sv.signon, ent->v.skin);

	// not on the server so no need to rough-round angles[1]????
	MSG_WriteCoord (&sv.signon, ent->v.origin[0], sv.protocol, sv.protocolflags);
	MSG_WriteAngle (&sv.signon, ent->v.angles[0], sv.protocol, sv.protocolflags, false);

	MSG_WriteCoord (&sv.signon, ent->v.origin[1], sv.protocol, sv.protocolflags);
	MSG_WriteAngle (&sv.signon, ent->v.angles[1], sv.protocol, sv.protocolflags, false);

	MSG_WriteCoord (&sv.signon, ent->v.origin[2], sv.protocol, sv.protocolflags);
	MSG_WriteAngle (&sv.signon, ent->v.angles[2], sv.protocol, sv.protocolflags, false);

	// johnfitz -- PROTOCOL_FITZQUAKE
	if (bits & B_ALPHA)
		MSG_WriteByte (&sv.signon, ent->alpha);

	// johnfitz

	// throw the entity away now
	ED_Free (ent);
}

// =============================================================================

/*
==============
PF_setspawnparms
==============
*/
void PF_setspawnparms (void)
{
	edict_t	*ent;
	int		i;
	client_t	*client;

	ent = G_EDICT (OFS_PARM0);
	i = GetNumberForEdict (ent);

	if (i < 1 || i > svs.maxclients)
		PR_RunError ("Entity is not a client");

	// copy spawn parms out of the client_t
	client = svs.clients + (i - 1);

	for (i = 0; i < NUM_SPAWN_PARMS; i++)
		(&pr_progs->global_struct->parm1)[i] = client->spawn_parms[i];
}

/*
==============
PF_changelevel
==============
*/
void PF_changelevel (void)
{
	char	*s;

	// make sure we don't issue two changelevels
	if (svs.changelevel_issued)
		return;

	svs.changelevel_issued = true;

	s = G_STRING (OFS_PARM0);
	Cbuf_AddText (va ("changelevel %s\n", s));
}

void PF_Fixme (void)
{
	PR_RunError ("unimplemented bulitin");
}


// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  start
/*
=================
PF_builtin_find

float builtin_find (string)
=================
*/
void PF_builtin_find (void)
{
	int		j;
	float	funcno;
	char	*funcname;

	funcno = 0;
	funcname = G_STRING(OFS_PARM0);

	// search function name
	for (j = 1; j < pr_progs->ebfs_numbuiltins; j++)
	{
		if ((pr_progs->ebfs_builtins[j].funcname) && (!(Q_strcasecmp (funcname, pr_progs->ebfs_builtins[j].funcname))))
		{
			break;	// found
		}
	}

	if (j < pr_progs->ebfs_numbuiltins)
	{
		funcno = pr_progs->ebfs_builtins[j].funcno;
	}

	G_FLOAT (OFS_RETURN) = funcno;
}
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  end


/*
=========
PF_strftime

string strftime(float uselocaltime, string[, string ...])
=========
*/
void PF_strftime (void)
{
	time_t t;

#if _MSC_VER >= 1400
	struct tm tm;
	int tmresult;
#else
	struct tm *tm;
#endif

	// because we don't have DP's temp string stuff we'll make these static instead
	static char result[STRINGTEMP_LENGTH];

	// bug 1 - this needs it's return value stored somewhere...
	char *fmt = PF_VarString (1);

	t = time (NULL);

#if _MSC_VER >= 1400
	// bug 2 so does this???
	// bug 3 - localtime_s returns zero if successful
	// bug 4 - so does gmtime_s
	if (G_FLOAT (OFS_PARM0))
		tmresult = localtime_s(&tm, &t);
	else tmresult = gmtime_s(&tm, &t);

	if (tmresult)
#else
	if (G_FLOAT (OFS_PARM0))
		tm = localtime(&t);
	else tm = gmtime(&t);

	if (!tm)
#endif
	{
		G_INT (OFS_RETURN) = 0;
		return;
	}

#if _MSC_VER >= 1400
	strftime(result, sizeof(result), fmt, &tm);
#else
	strftime(result, sizeof(result), fmt, tm);
#endif

	G_INT (OFS_RETURN) = PR_SetString (result);
}


void PF_tracebox (void)
{
	float   *v1, *v2, *mins, *maxs;
	trace_t   trace;
	int      nomonsters;
	edict_t   *ent;

	v1 = G_VECTOR(OFS_PARM0);

	mins = G_VECTOR(OFS_PARM1);
	maxs = G_VECTOR(OFS_PARM2);

	v2 = G_VECTOR(OFS_PARM3);

	nomonsters = G_FLOAT (OFS_PARM4);

	ent = G_EDICT(OFS_PARM5);

	trace = SV_Move (v1, mins, maxs, v2, nomonsters, ent);

	pr_progs->global_struct->trace_allsolid = trace.allsolid;
	pr_progs->global_struct->trace_startsolid = trace.startsolid;
	pr_progs->global_struct->trace_fraction = trace.fraction;
	pr_progs->global_struct->trace_inwater = trace.inwater;
	pr_progs->global_struct->trace_inopen = trace.inopen;

	VectorCopy (trace.endpos, pr_progs->global_struct->trace_endpos);
	VectorCopy (trace.plane.normal, pr_progs->global_struct->trace_plane_normal);

	pr_progs->global_struct->trace_plane_dist =  trace.plane.dist;

	if (trace.ent)
		pr_progs->global_struct->trace_ent = EDICT_TO_PROG (trace.ent);
	else pr_progs->global_struct->trace_ent = EDICT_TO_PROG (pr_progs->edicts[0]);
}


void PF_te_particlerain (void)
{
	if (G_FLOAT (OFS_PARM3) < 1) return;

	MSG_WriteByte (&sv.datagram, svc_temp_entity);
	MSG_WriteByte (&sv.datagram, TE_PARTICLERAIN);

	// min
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM0)[0], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM0)[1], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM0)[2], sv.protocol, sv.protocolflags);

	// max
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM1)[0], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM1)[1], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM1)[2], sv.protocol, sv.protocolflags);

	// velocity
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM2)[0], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM2)[1], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM2)[2], sv.protocol, sv.protocolflags);

	// count - already tested for < 1 above
	MSG_WriteShort (&sv.datagram, G_FLOAT (OFS_PARM3) > 65535 ? 65535 : G_FLOAT (OFS_PARM3));

	// color
	MSG_WriteByte (&sv.datagram, G_FLOAT (OFS_PARM4));
}


void PF_te_particlesnow (void)
{
	if (G_FLOAT (OFS_PARM3) < 1) return;

	MSG_WriteByte (&sv.datagram, svc_temp_entity);
	MSG_WriteByte (&sv.datagram, TE_PARTICLESNOW);

	// min
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM0)[0], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM0)[1], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM0)[2], sv.protocol, sv.protocolflags);

	// max
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM1)[0], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM1)[1], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM1)[2], sv.protocol, sv.protocolflags);

	// velocity
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM2)[0], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM2)[1], sv.protocol, sv.protocolflags);
	MSG_WriteCoord (&sv.datagram, G_VECTOR (OFS_PARM2)[2], sv.protocol, sv.protocolflags);

	// count - already tested for < 1 above
	MSG_WriteShort (&sv.datagram, G_FLOAT (OFS_PARM3) > 65535 ? 65535 : G_FLOAT (OFS_PARM3));

	// color
	MSG_WriteByte (&sv.datagram, G_FLOAT (OFS_PARM4));
}


// 2001-09-20 QuakeC string manipulation by FrikaC/Maddes  start

// mh - some day we'll get these OUT of the fucking zone, but for now they might break so
// until we have a working implementation that uses them we'll leave it be...
/*
=================
PF_strzone

string strzone (string)
=================
*/
void PF_strzone (void)
{
	char *m, *p;

	m = G_STRING (OFS_PARM0);
	p = calloc (strlen (m) + 1, 1);

	strcpy (p, m);

	G_INT(OFS_RETURN) = PR_SetString (p);
}


/*
=================
PF_strunzone

string strunzone (string)
=================
*/
void PF_strunzone (void)
{
	// the compiler freaks if I pass this directly into free...
	void *blah = (void *) (G_STRING (OFS_PARM0));

	free (blah);
	G_INT (OFS_PARM0) = OFS_NULL; // empty the def
}


/*
=================
PF_strlen

float strlen (string)
=================
*/
void PF_strlen (void)
{
	char *p = G_STRING(OFS_PARM0);

	G_FLOAT (OFS_RETURN) = strlen(p);
}


/*
=================
PF_strcat

string strcat (string, string)
=================
*/
void PF_strcat (void)
{
	char *s1, *s2;
	int		maxlen;	// 2001-10-25 Enhanced temp string handling by Maddes
	char	*tempstr = PR_GetTempString ();

	s1 = G_STRING(OFS_PARM0);
	s2 = PF_VarString(1);

// 2001-10-25 Enhanced temp string handling by Maddes  start
	tempstr[0] = 0;

	if (strlen(s1) < PR_MAX_TEMPSTRING)
	{
		strcpy(tempstr, s1);
	}
	else
	{
		strncpy(tempstr, s1, PR_MAX_TEMPSTRING);
		tempstr[PR_MAX_TEMPSTRING-1] = 0;
	}

	maxlen = PR_MAX_TEMPSTRING - strlen(tempstr) - 1;	// -1 is EndOfString

	if (maxlen > 0)
	{
		if (maxlen > strlen(s2))
		{
			strcat (tempstr, s2);
		}
		else
		{
			strncat (tempstr, s2, maxlen);
			tempstr[PR_MAX_TEMPSTRING-1] = 0;
		}
	}
// 2001-10-25 Enhanced temp string handling by Maddes  end

	G_INT(OFS_RETURN) = PR_SetString (tempstr);
}

/*FTE_STRINGS #222
added purely because csqctest likes to use it for the hud
*/
void PF_str2chr (void)
{
	char *instr = G_STRING(OFS_PARM0);
	int ofs = (pr_progs->prargc>1)?G_FLOAT(OFS_PARM1):0;

	if (ofs < 0)
		ofs = strlen(instr)+ofs;

	if (ofs && (ofs < 0 || ofs > strlen(instr)))
		G_FLOAT(OFS_RETURN) = '\0';
	else
		G_FLOAT(OFS_RETURN) = instr[ofs];
}


/*
=================
PF_substring

string substring (string, float, float)
=================
*/
void PF_substring (void)
{
	int		offset, length;
	int		maxoffset;		// 2001-10-25 Enhanced temp string handling by Maddes
	char	*p;
	char	*tempstr = PR_GetTempString ();

	p = G_STRING(OFS_PARM0);
	offset = (int)G_FLOAT (OFS_PARM1); // for some reason, Quake doesn't like G_INT
	length = (int)G_FLOAT (OFS_PARM2);

	// cap values
	maxoffset = strlen(p);

	if (offset > maxoffset)
	{
		offset = maxoffset;
	}

	if (offset < 0)
		offset = 0;

// 2001-10-25 Enhanced temp string handling by Maddes  start
	if (length >= PR_MAX_TEMPSTRING)
		length = PR_MAX_TEMPSTRING-1;
// 2001-10-25 Enhanced temp string handling by Maddes  end

	if (length < 0)
		length = 0;

	p += offset;
	strncpy(tempstr, p, length);
	tempstr[length]=0;

	G_INT(OFS_RETURN) = PR_SetString (tempstr);
}


/*
=================
PF_stof

float stof (string)
=================
*/

// thanks Zoid, taken from QuakeWorld
void PF_stof (void)
{
	char	*s;

	s = G_STRING(OFS_PARM0);
	G_FLOAT (OFS_RETURN) = atof(s);
}


/*
=================
PF_stov

vector stov (string)
=================
*/
void PF_stov (void)
{
	char *v;
	int i;
	vec3_t d;

	v = G_STRING(OFS_PARM0);

	for (i=0; i<3; i++)
	{
		while(v && (v[0] == ' ' || v[0] == '\'')) // skip unneeded data
			v++;

		d[i] = atof(v);

		while (v && v[0] != ' ') // skip to next space
			v++;
	}
	VectorCopy (d, G_VECTOR(OFS_RETURN));
}
// 2001-09-20 QuakeC string manipulation by FrikaC/Maddes  end


// 2001-09-20 QuakeC file access by FrikaC/Maddes  start
/*
=================
PF_fopen

float fopen (string,float)
=================
*/
void PF_fopen (void)
{
	char *p = G_STRING(OFS_PARM0);
	int fmode = G_FLOAT (OFS_PARM1);
	int h = 0;

	switch (fmode)
	{
	case 0: // read
		Sys_FileOpenRead (va("%s/%s",com_gamedir, p), &h);
		break;

	case 1: // append -- sane version
		h = Sys_FileOpenAppend (va("%s/%s", com_gamedir, p));
		break;

	default: // write
		h = Sys_FileOpenWrite (va("%s/%s", com_gamedir, p));
		break;
	}

	// always common now
	G_FLOAT (OFS_RETURN) = (float) h;
}


/*
=================
PF_fclose

void fclose (float)
=================
*/
void PF_fclose (void)
{
	int h = (int)G_FLOAT (OFS_PARM0);

	Sys_FileClose(h);
}


/*
=================
PF_fgets

string fgets (float)
=================
*/
void PF_fgets (void)
{
	// reads one line (up to a \n) into a string
	int		h;
	int		i;
	int		count;
	char	buffer;
	char	*tempstr = PR_GetTempString ();

	h = (int)G_FLOAT (OFS_PARM0);
	count = Sys_FileRead(h, &buffer, 1);

	if (count && buffer == '\r')	// carriage return
	{
		count = Sys_FileRead(h, &buffer, 1);	// skip
	}

	if (!count)	// EndOfFile
	{
		G_INT(OFS_RETURN) = OFS_NULL;	// void string
		return;
	}

	i = 0;

	while (count && buffer != '\n')
	{
		if (i < PR_MAX_TEMPSTRING-1)	// no place for character in temp string
		{
			tempstr[i++] = buffer;
		}

		// read next character
		count = Sys_FileRead(h, &buffer, 1);

		if (count && buffer == '\r')	// carriage return
		{
			count = Sys_FileRead(h, &buffer, 1);	// skip
		}
	}

	tempstr[i] = 0;

	G_INT(OFS_RETURN) = PR_SetString (tempstr);
}


/*
=================
PF_fputs

void fputs (float,string)
=================
*/
void PF_fputs (void)
{
	// writes to file, like bprint
	float handle = G_FLOAT (OFS_PARM0);
	char *str = PF_VarString(1);

	Sys_FileWrite (handle, str, strlen(str));
}
// 2001-09-20 QuakeC file access by FrikaC/Maddes  end


// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  start

// for builtin function definitions see Quake Standards Group at http://www.quakesrc.org/

// server side builtins
ebfs_builtin_t sv_ebfs_builtins[] =
{
	{   0, NULL, PF_Fixme },		// has to be first entry as it is needed for initialization in PR_LoadProgs ()
	{   1, "makevectors", PF_makevectors },	// void(entity e)	makevectors 		= #1;
	{   2, "setorigin", PF_setorigin },		// void(entity e, vector o) setorigin	= #2;
	{   3, "setmodel", PF_setmodel },		// void(entity e, string m) setmodel	= #3;
	{   4, "setsize", PF_setsize },			// void(entity e, vector min, vector max) setsize = #4;
// 	{   5, "fixme", PF_Fixme },				// void(entity e, vector min, vector max) setabssize = #5;
	{   6, "break", PF_break },				// void () break						= #6;
	{   7, "random", PF_random },			// float () random						= #7;
	{   8, "sound", PF_sound },				// void(entity e, float chan, string samp) sound = #8;
	{   9, "normalize", PF_normalize },		// vector(vector v) normalize			= #9;
	{  10, "error", PF_error },				// void(string e) error				= #10;
	{  11, "objerror", PF_objerror },		// void(string e) objerror				= #11;
	{  12, "vlen", PF_vlen },				// float(vector v) vlen				= #12;
	{  13, "vectoyaw", PF_vectoyaw },		// float(vector v) vectoyaw		= #13;
	{  14, "spawn", PF_Spawn },				// entity () spawn						= #14;
	{  15, "remove", PF_Remove },			// void(entity e) remove				= #15;
	{  16, "traceline", PF_traceline },		// float(vector v1, vector v2, float tryents) traceline = #16;
	{  17, "checkclient", PF_checkclient },	// entity () clientlist					= #17;
	{  18, "find", PF_Find },				// entity(entity start, .string fld, string match) find = #18;
	{  19, "precache_sound", PF_precache_sound },	// void(string s) precache_sound		= #19;
	{  20, "precache_model", PF_precache_model },	// void(string s) precache_model		= #20;
	{  21, "stuffcmd", PF_stuffcmd },		// void(entity client, string s)stuffcmd = #21;
	{  22, "findradius", PF_findradius },	// entity(vector org, float rad) findradius = #22;
	{  23, "bprint", PF_bprint },			// void(string s) bprint				= #23;
	{  24, "sprint", PF_sprint },			// void(entity client, string s) sprint = #24;
	{  25, "dprint", PF_dprint },			// void(string s) dprint				= #25;
	{  26, "ftos", PF_ftos },				// void(string s) ftos				= #26;
	{  27, "vtos", PF_vtos },				// void(string s) vtos				= #27;
	{  28, "coredump", PF_coredump },
	{  29, "traceon", PF_traceon },
	{  30, "traceoff", PF_traceoff },
	{  31, "eprint", PF_eprint },			// void(entity e) debug print an entire entity
	{  32, "walkmove", PF_walkmove },		// float(float yaw, float dist) walkmove
// 	{  33, "fixme", PF_Fixme },				// float(float yaw, float dist) walkmove
	{  34, "droptofloor", PF_droptofloor },
	{  35, "lightstyle", PF_lightstyle },
	{  36, "rint", PF_rint },
	{  37, "floor", PF_floor },
	{  38, "ceil", PF_ceil },
// 	{  39, "fixme", PF_Fixme },
	{  40, "checkbottom", PF_checkbottom },
	{  41, "pointcontents", PF_pointcontents },
// 	{  42, "fixme", PF_Fixme },
	{  43, "fabs", PF_fabs },
	{  44, "aim", PF_aim },
	{  45, "cvar", PF_cvar },
	{  46, "localcmd", PF_localcmd },
	{  47, "nextent", PF_nextent },
	{  48, "particle", PF_particle },
	{  49, "ChangeYaw", PF_changeyaw },
// 	{  50, "fixme", PF_Fixme },
	{  51, "vectoangles", PF_vectoangles },
	{  52, "WriteByte", PF_WriteByte },
	{  53, "WriteChar", PF_WriteChar },
	{  54, "WriteShort", PF_WriteShort },
	{  55, "WriteLong", PF_WriteLong },
	{  56, "WriteCoord", PF_WriteCoord },
	{  57, "WriteAngle", PF_WriteAngle },
	{  58, "WriteString", PF_WriteString },
	{  59, "WriteEntity", PF_WriteEntity },

	{  60, "sin", PF_sin },
	{  61, "cos", PF_cos },
	{  62, "sqrt", PF_sqrt },
	/*
	---QUAKE2---
	{  63, "changepitch", PF_changepitch },
	{  64, "TraceToss", PF_TraceToss },
	{  65, "etos", PF_etos },
	{  66, "WaterMove", PF_WaterMove },
	---QUAKE2---
	*/

	{  67, "movetogoal", SV_MoveToGoal },
	{  68, "precache_file", PF_precache_file },
	{  69, "makestatic", PF_makestatic },
	{  70, "changelevel", PF_changelevel },
// 	{  71, "fixme", PF_Fixme },
	{  72, "cvar_set", PF_cvar_set },
	{  73, "centerprint", PF_centerprint },
	{  74, "ambientsound", PF_ambientsound },
	{  75, "precache_model2", PF_precache_model },
	{  76, "precache_sound2", PF_precache_sound },	// precache_sound2 is different only for qcc
	{  77, "precache_file2", PF_precache_file },
	{  78, "setspawnparms", PF_setspawnparms },
// 	{  81, "stof", PF_stof },	// 2001-09-20 QuakeC string manipulation by FrikaC/Maddes
// 2001-11-15 DarkPlaces general builtin functions by Lord Havoc  start

// not implemented yet
	{  90, "tracebox", PF_tracebox },
	{  91, "randomvec", PF_randomvec },
/*
	{  92, "getlight", PF_GetLight },	// not implemented yet
	{  93, "cvar_create", PF_cvar_create },		// 2001-09-18 New BuiltIn Function: cvar_create () by Maddes
	{  94, "fmin", PF_fmin },
	{  95, "fmax", PF_fmax },
	{  96, "fbound", PF_fbound },*/
	{  97, "fpow", PF_fpow },
/*	{  98, "findfloat", PF_FindFloat },
*/
	{ PR_DEFAULT_FUNCNO_EXTENSION_FIND, "extension_find", PF_extension_find },	// 2001-10-20 Extension System by Lord Havoc/Maddes
/*
	{   0, "registercvar", PF_cvar_create },	// 0 indicates that this entry is just for remapping (because of name change)
	{   0, "checkextension", PF_extension_find },
*/
// 2001-11-15 DarkPlaces general builtin functions by Lord Havoc  end

	{ PR_DEFAULT_FUNCNO_BUILTIN_FIND, "builtin_find", PF_builtin_find },	// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes

// not implemented yet
/*
	{ 101, "cmd_find", PF_cmd_find },		// 2001-09-16 New BuiltIn Function: cmd_find () by Maddes
	{ 102, "cvar_find", PF_cvar_find },		// 2001-09-16 New BuiltIn Function: cvar_find () by Maddes
	{ 103, "cvar_string", PF_cvar_string },	// 2001-09-16 New BuiltIn Function: cvar_string () by Maddes
	{ 105, "cvar_free", PF_cvar_free },		// 2001-09-18 New BuiltIn Function: cvar_free () by Maddes
	{ 106, "NVS_InitSVCMsg", PF_NVS_InitSVCMsg },	// 2000-05-02 NVS SVC by Maddes
	{ 107, "WriteFloat", PF_WriteFloat },	// 2001-09-16 New BuiltIn Function: WriteFloat () by Maddes
	{ 108, "etof", PF_etof },	// 2001-09-25 New BuiltIn Function: etof () by Maddes
	{ 109, "ftoe", PF_ftoe },	// 2001-09-25 New BuiltIn Function: ftoe () by Maddes
*/
// 2001-09-20 QuakeC file access by FrikaC/Maddes  start
// not implemented yet
	{ 110, "fopen", PF_fopen },
	{ 111, "fclose", PF_fclose },
	{ 112, "fgets", PF_fgets },
	{ 113, "fputs", PF_fputs },
	{   0, "open", PF_fopen },		// 0 indicates that this entry is just for remapping (because of name and number change)
	{   0, "close", PF_fclose },
	{   0, "read", PF_fgets },
	{   0, "write", PF_fputs },
// 2001-09-20 QuakeC file access by FrikaC/Maddes  end

// 2001-09-20 QuakeC string manipulation by FrikaC/Maddes  start
// not implemented yet
	{ 114, "strlen", PF_strlen },
	{ 115, "strcat", PF_strcat },
	{ 116, "substring", PF_substring },
	{ 117, "stov", PF_stov },
	{ 118, "strzone", PF_strzone },
	{ 119, "strunzone", PF_strunzone },
	{   0, "zone", PF_strzone },		// 0 indicates that this entry is just for remapping (because of name and number change)
	{   0, "unzone", PF_strunzone },
// 2001-09-20 QuakeC string manipulation by FrikaC/Maddes  end

// 2001-11-15 DarkPlaces general builtin functions by Lord Havoc  start
// not implemented yet
/*
	{ 400, "copyentity", PF_... },
	{ 401, "setcolor", PF_... },
	{ 402, "findchain", PF_... },
	{ 403, "findchainfloat", PF_... },
	{ 404, "effect", PF_... },
	{ 405, "te_blood", PF_... },
	{ 406, "te_bloodshower", PF_... },
	{ 407, "te_explosionrgb", PF_... },
	{ 408, "te_particlecube", PF_... },
	*/
	{409, "te_particlerain", PF_te_particlerain},
	{410, "te_particlesnow", PF_te_particlesnow},
	/*
	{ 411, "te_spark", PF_... },
	{ 412, "te_gunshotquad", PF_... },
	{ 413, "te_spikequad", PF_... },
	{ 414, "te_superspikequad", PF_... },
	{ 415, "te_explosionquad", PF_... },
	{ 416, "te_smallflash", PF_... },
	{ 417, "te_customflash", PF_... },
	{ 418, "te_gunshot", PF_... },
	{ 419, "te_spike", PF_... },
	{ 420, "te_superspike", PF_... },
	{ 421, "te_explosion", PF_... },
	{ 422, "te_tarexplosion", PF_... },
	{ 423, "te_wizspike", PF_... },
	{ 424, "te_knightspike", PF_... },
	{ 425, "te_lavasplash", PF_... },
	{ 426, "te_teleport", PF_... },
	{ 427, "te_explosion2", PF_... },
	{ 428, "te_lightning1", PF_... },
	{ 429, "te_lightning2", PF_... },
	{ 430, "te_lightning3", PF_... },
	{ 431, "te_beam", PF_... },
	{ 432, "vectorvectors", PF_... },
*/
	{ 478, "strftime", PF_strftime},
// 2001-11-15 DarkPlaces general builtin functions by Lord Havoc  end

	{ 666, "getrmqeffectsversion", PF_GetRMQEffectsVersion }
};
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  end


/*
================================================================================================================

			CSQC CRAP

	VM CSQC builtins are gathered here for convenience, clarity and ease of access.

	Some of these may make calls into other parts of the engine, and some of these may be implementation
	dependent (as in, dependent on how things are coded in RMQ) so be careful when copy-pasting!

================================================================================================================
*/

/*
TODO: after the client has connected to the server, the server SHOULD send a hash+length or so of the module the server 'permits'. Other modules SHOULD not be loaded. A downloading mechanism is recommended...
This provides cheat prevention, and peace of mind regarding people making csqc cheats (they need to mod their engine, thus csqc availability is no new avenue to cheats in itself).
(FTE uses an md4 32bit hash, DP uses CCIT CRC and length).
If no hash is received by the client, the engine should refuse to load the csprogs.
Strictly speaking the behaviour mentioned above is optional, particuarly with demos, but it is fairly highly recommended.
(FTE caches the downloaded progs to a subdirectory with the hash hex-encoded in the file name, when it says its using 'csprogs.dat', only the server reads that, the client uses only the cache (unless it has no cache, and the local csprogs does match)).

The csprogs is loaded only after precache lists are received by the client, but logically before anything is loaded.
After the cprogs has been validated and loaded, the csqc entities are initialised. specifically, a null/world entity is spawned.
The engine then calls a function with the prototype:
void(float csqcspecversion, string enginestring, float engineversion) CSQC_Init;
if it exists.
This function may call precache_model (possibly on modelnames that match the ones from the server) etc
This function may call functions to redraw the screen (a loading screen), so the csqc system must be running, even if models from the server are not yet loaded.
FIXME: this was the idea, yet there is no 'swapbuffers' call (which would presumably also ensure that csqc-precached models are loaded before returning).

The client continues loading its stuff, models from the server, models that were precached from csqc.

At some point, the client sets up world.modelindex, world.model, world.solid at a minimum.
At some later point, the client calls void () CSQC_WorldLoaded.
Once this call happens, all models and sounds from the server are fully loaded. The csqc may have more to add though.
For the duration of this call, getentitytoken may be called to parse the map's entities, as loaded from the bsp.
After CSQC_WorldLoaded returns, the csqc's world entity can no longer be written to.

The csprogs is now running. csqc should avoid further precaches, but is not prevented from doing so.

Each frame:
Basic client frame logic works like this:
Receives updates.
Calculate simulation/lerp time.
Set csqcglobal.time = simulationtime;
Set csqcglobal.cstime = client realtime since start of map (eg: Sys_DoubleTime ());
Set csqcglobal.servercommandframe = the input log frame that was last acknowledged by the server.
Set csqcglobal.clientcommandframe = the input log frame that will be generated next.
 Engine calls CSQC_UpdateView.
 CSQC optionally calls various 2d functions.
 CSQC optionally:
  calls ClearScene
  calls SetView
  calls AddEntity/AddEntities
  calls RenderScene
 CSQC optionally calls various 2d functions.
 CSQC optionally draws a second, third, etc scene.
 CSQC returns.
 Engine draws console+centerprint+menu.
 Engine swaps buffers.

For each packet sent to the server:
Receive user input.
Engine fills input_foo globals.
Engine calls CSQC_Input_Frame.
Engine copies input_foo globals into an input-frame log (log = &entry[clientcommandframe++ % entries]).
Engine sends most recent log entry.

Without CSQC, the engine effectively calls ClearScene through to RenderScene automagically.

SIDENOTE: it has been mentioned that CSQC is hard to get into due to it requiring a certain set of calls. namely, ClearScene, Setview(multiple times), AddEntities, RenderScene, for a basic empty mod.

Delta/Engine entities need a special mention.
When using CSQC, they are by default not drawn.
They are added to the scene only when AddEntities(mask&1) is called.
Somewhere inside either AddEntities&1, or at the start of every frame, DeltaListen entities are handled, by copying them to csqc entities. csqc can set a drawmask on these ents to get them drawn.

Interpolation: ents have a frame(1), a frame2, and a lerpfrac field (aka: frame2ness). These fields are filled in by the qc (the exception is delta entities, which the client fills the fields for each frame)
additionally, frame1time and frame2time fields exist, which control time into framegroup animations. these fields can be written to adjust framegroup animation rates.
There is a flag to make frameXtime a timestamp and thus autoanimating.

While parsing network messages, there are certain hooks into csqc code.
CSQC can hook stuffcmds, prints, centerprints, and arguably sounds (sounds are unreliable, as they might be attached to an entity that is not known, thus cannot be relied upon to be useful).

The CSQC networking protocol adds csqc entities. The engine handles which entities are sent, and which flags need to be sent, but does not have any say over the data associated with each entity.
Additionally, there are at least 128 stats (the first 32 are reserved for engine/non-csqc use). FTE greatly changes its stat system to provide more utility to mods, but its not guarenteed to be compatible with DP mods that report ints where they provide floats.
Additionally, support for the client input frame acknowledgement is required.

The input log is what permits prediction.
if the entity state is reset to servercommandframe, then applying each input frame starting at servercommandframe and up to but not including clientinputframe will provide accurate prediction if the server and client agree on everything else.
The server doesn't care for the clients prediction on the condition that it runs the client frames in their entirety, in sequence, and reports the most recent frame applied to that player's entity's state, without server timings affecting the movement times.

precaches:
server precaches are positive
clientside-only precaches are negative.
a precache which is already on the server should return a positive index
a mid-map precache from the server can result in two valid indexes for the same model/sound (as could a buggy server, I suppose).

2d stuff:
2d stuff is fairly simple. There's an optional 'flags' field in most places. this is a relic of menuqc, and permits additive/modulatex2 blends instead. its only real use is for nexuiz's pre-csqc hud.
*/



void V_CalcRefdef (void);
void V_CalcIntermissionRefdef (void);

/*set view properties to their defaults (view org, fov, view size, r_refdef type stuff), remove all ents from the view (beams are hidden, engine ents are hidden.). defaults are as the engine would normally have*/
void PFCS_R_ClearScene (void)
{
#if TODO3D
	/*Forget all entities*/
	cl_numvisedicts = 0;
	V_CalcRefdef (); /*TODO: this is dependant upon framerate, and thus will be dependant upon number of invocations per frame
					it also expects the player's local entity to already be updated
					*/
#endif

	if (cl.intermission)
		V_CalcIntermissionRefdef ();
	else
		V_CalcRefdef ();
}

/*adds groups of entities to the view. there are some special cases
loop through all ents, if the mask matches its drawmask, the ent is added to the scene (see R_AddEntity)
if mask&1, call CL_LinkEntities. Also adds/updates beams(aka: tents).
if mask&2, the engine's viewmodel will be drawn.*/
void PFCS_R_AddEntities (void)
{
#if TODO3D
	if (mask & 1)
	{
		CL_RelinkEntities ();
		CL_UpdateTEnts ();
	}
	if (mask & 2)
	{
		CL_LinkViewModel ();
	}
	foreach(csqc edict)
	{
		if (ed->free)
			continue;

		if (ed->v.drawmask & mask)
		{
			entity_t *ent;
			if (ent->v.predraw)
			{
				globals->self = ed;
				PR_ExecuteProgram(ed->v.predraw);
				/*if it freed itself, or it flaged itself for not being added automatically, we need to give up*/
				if (ed->free || (int)ent->xv->renderflags & CSQCRF_NOAUTOADD)
					continue;
			}
			/*okay, it needs to be visible, make a new ent for it and fill in the fields*/
			ent = CL_NewTempEntity ();
			fillinentfromed(ent, ed);

			if (cl_numvisedicts < MAX_VISEDICTS)
			{
				cl_visedicts[cl_numvisedicts] = ent;
				cl_numvisedicts++;
			}
		}
	}

#endif
}

/*adds a single entity to the view (distinction made between (vm)edict and (r)entity*/
void PFCS_R_AddEntity (void)
{
#if TODO3D
	edict_t	*ed = (edict_t*)G_EDICT(OFS_PARM0);
	entity_t *ent = CL_NewTempEntity ();
	fillinentfromed(ent, ed);

	if (cl_numvisedicts < MAX_VISEDICTS)
	{
		cl_visedicts[cl_numvisedicts] = ent;
		cl_numvisedicts++;
	}
#endif
}

/*sets a view property, second argument is a variant, qc may need multiple forms of the builtin*/
void PFCS_R_SetView (void)
{
	int parametertype = G_FLOAT(OFS_PARM0);
	float *p = G_VECTOR(OFS_PARM1);

	/*TODO3D*/
	switch(parametertype)
	{
	case VF_MIN:
		r_refdef.vrectset = true;
		r_refdef.vrect.x = p[0];
		r_refdef.vrect.y = p[1];
		break;
	case VF_MIN_X:
		r_refdef.vrectset = true;
		r_refdef.vrect.x = p[0];
		break;
	case VF_MIN_Y:
		r_refdef.vrectset = true;
		r_refdef.vrect.y = p[0];
		break;
	case VF_SIZE:	// (vector) (viewport size)
		r_refdef.vrectset = true;
		r_refdef.vrect.width = p[0];
		r_refdef.vrect.height = p[1];
		break;
	case VF_SIZE_X:	// (float)
		r_refdef.vrectset = true;
		r_refdef.vrect.width = p[0];
		break;
	case VF_SIZE_Y:	// (float)
		r_refdef.vrectset = true;
		r_refdef.vrect.height = p[0];
		break;
	case VF_VIEWPORT:	// (vector, vector)
		r_refdef.vrectset = true;
		r_refdef.vrect.x = p[0];
		r_refdef.vrect.y = p[1];
		r_refdef.vrect.width = p[3];
		r_refdef.vrect.height = p[4];
		break;
	case VF_FOV:	// (vector)
		r_refdef.fov_x = p[0];
		r_refdef.fov_y = p[1];
		break;
	case VF_FOVX:	// (float)
		r_refdef.fov_x = p[0];
		break;
	case VF_FOVY:	// (float)
		r_refdef.fov_y = p[0];
		break;
	case VF_ORIGIN:	// (vector)
		VectorCopy(p, r_refdef.vieworg);
		break;
	case VF_ORIGIN_X:	// (float)
		r_refdef.vieworg[0] = p[0];
		break;
	case VF_ORIGIN_Y:	// (float)
		r_refdef.vieworg[1] = p[0];
		break;
	case VF_ORIGIN_Z:	// (float)
		r_refdef.vieworg[2] = p[0];
		break;
	case VF_ANGLES:	// (vector)
		VectorCopy(p, r_refdef.viewangles);
		break;
	case VF_ANGLES_X:	// (float)
		r_refdef.viewangles[0] = p[0];
		break;
	case VF_ANGLES_Y:	// (float)
		r_refdef.viewangles[1] = p[0];
		break;
	case VF_ANGLES_Z:	// (float)
		r_refdef.viewangles[2] = p[0];
		break;

	case VF_DRAWWORLD:	// (float)	// actually world model and sky
		// this is bogus client-side cheat protection
		break;
	case VF_DRAWENGINESBAR:	// (float)
		cl.qc.hidehud = !p[0];
		break;
	case VF_DRAWCROSSHAIR:	// (float)
		cl.qc.hidecrosshair = !p[0];
		break;

	/*these ones are DP hacks, and affect the player instead of the view (what did you think the V in VF stood for...)*/
	case VF_CL_VIEWANGLES:
		VectorCopy(p, cl.viewangles);
		break;
	case VF_CL_VIEWANGLES_X:	// (float)
		cl.viewangles[0] = p[0];
		break;
	case VF_CL_VIEWANGLES_Y:	// (float)
		cl.viewangles[1] = p[0];
		break;
	case VF_CL_VIEWANGLES_Z:	// (float)
		cl.viewangles[2] = p[0];
		break;

	case VF_PERSPECTIVE:
		/*the 200+ range are fte extensions, consistant with builtins*/
		break;
	}
}

/*draws the currently submitted entities and probably but not always the world. if world is drawn, particles are also drawn, otherwise not*/
void PFCS_R_RenderScene (void)
{
	/*draw whatever was pushed to the render lists*/
	R_RenderView ();

	/*and reset the screen for working 2d drawing*/
	GL_Set2D ();
}

/*adds a dynamic light to the scene*/
void PFCS_R_AddDynamicLight (void)
{
	float *org = G_VECTOR(OFS_PARM0);
	float radius = G_FLOAT(OFS_PARM1);
	float *rgb = G_VECTOR(OFS_PARM2);
	/*optional: int style = G_FLOAT(OFS_PARM3);*/
	/*optional: char *cubemapname = G_STRING(OFS_PARM4);*/
	/*optional: float pflags = G_FLOAT(OFS_PARM5);*/
	dlight_t *dl;

	dl = CL_AllocDlight (0);
	VectorCopy (org,  dl->origin);
	dl->origin[2] += 16;
	dl->radius = radius;
	dl->die = 0; /*last only one rendering frame*/
	/*TODO*/
}

/*extension, acts a bit like glBegin. Polygons are stored for later rendering. You can directly write them to a vbo if you want.*/
void PFCS_R_BeginPolygon (void)
{
	char *texturename = G_STRING(OFS_PARM0);
	/*Optional TODO*/
}

/*extension, acts like glColor+glTexCoord+glVertex*/
void PFCS_R_PolygonVertex (void)
{
	float *xyz = G_VECTOR(OFS_PARM0);
	float *st = G_VECTOR(OFS_PARM1);
	float *rgb = G_VECTOR(OFS_PARM2);
	float alpha = G_FLOAT(OFS_PARM3);
	/*Optional TODO*/
}

/*extension, acts like glEnd*/
void PFCS_R_EndPolygon (void)
{
	/*triangulate polygon if needed*/
	/*Optional TODO*/
}

/*converts a 2d coord into a 3d coord according to the currently set up (R_SetView) view matrix+projection (2d coord has depth of 0 - 1 (1 might be infinite), 2d coords are 0-vid.conwidth/0-vid.conheight)*/
void PFCS_cs_unproject (void)
{
	float *in = G_VECTOR(OFS_PARM0);
	float *out = G_VECTOR(OFS_RETURN);
	/*TODO3D*/
}

/*converts a 3d coord into a 2d screen coord according to the currently set up (R_SetView) view matrix+projection (2d coord has depth of 0 - 1 (1 might be infinite, in which case clamp it to an epsilon closer to avoid NANs))*/
void PFCS_cs_project (void)
{
	float *in = G_VECTOR(OFS_PARM0);
	float *out = G_VECTOR(OFS_RETURN);
	/*TODO3D*/
}

/*draw a line between two points, possibly diagonally, final result is unreliable and only for diagram usage, so this is a low priority implementation...*/
void PFCS_drawline (void)
{
	float width = G_FLOAT(OFS_PARM0);
	float *rgb = G_VECTOR(OFS_PARM1);
	float alpha = G_FLOAT(OFS_PARM2);
	//Optional float flags = G_FLOAT(OFS_PARM3);
	float *pos1 = G_VECTOR(OFS_PARM4);
	float *pos2 = G_VECTOR(OFS_PARM5);
	/*Optional TODO*/
}

/*checks to see if an image is currently loaded, can just return a fixed value*/
void PFCS_iscachedpic (void)
{
	char *pic = G_STRING(OFS_PARM0);

	/*Lazy TODO2D*/
	G_FLOAT(OFS_RETURN) = 1;
}

/*ensures that an image is loaded, loading if its not*/
void PFCS_precache_pic (void)
{
	char *pic = G_STRING(OFS_PARM0);
	float fromwad = (pr_progs->prargc > 1)?G_FLOAT(OFS_PARM1):false;

	/*our cachepic checks wads implicitly*/
	Draw_CachePic(pic);
}

/*returns the images 'original' size. lmp replacements return the size of the lmp, not the replacement, if there is no original/lmp then returns the actual size*/
void PFCS_draw_getimagesize (void)
{
	char *pic = G_STRING(OFS_PARM0);

	float *ret = G_VECTOR(OFS_RETURN);

	ret[2] = 0;

	/*TODO2D*/
	ret[0] = 0;
	ret[1] = 0;
}

/*unloads a picture. doesn't have to be implemented*/
void PFCS_freepic (void)
{
	char *pic = G_STRING(OFS_PARM0);
	/*Optional TODO2D*/
}

/*the blend mode argument is an unnamed DP extension, and isn't needed for software*/
void PFCS_drawcharacter (void)
{
	float *pos = G_VECTOR(OFS_PARM0);
	int chara = G_FLOAT(OFS_PARM1);
	float *size = G_VECTOR(OFS_PARM2);
	float *rgb = G_VECTOR(OFS_PARM3);
	float alpha = G_FLOAT(OFS_PARM4);
	//Optional TODO: int flag = G_FLOAT(OFS_PARM5);
	byte rgba[4] = {rgb[0] * 255,
					rgb[1] * 255,
					rgb[2] * 255,
					alpha * 255};
	/*TODO2D: size is discarded (size is pixels, I wouldn't worry for software rendering anyway)*/
	Draw_Character(pos[0], pos[1], chara);
	Draw_ColoredCharacter(pos[0], pos[1], chara, *(unsigned int*)rgba);
}

/*see drawchar. does not parse ^4 etc markup*/
void PFCS_drawstring (void)
{
	float *pos = G_VECTOR(OFS_PARM0);
	char *text = G_STRING(OFS_PARM1);
	float *size = G_VECTOR(OFS_PARM2);
	float *rgb = G_VECTOR(OFS_PARM3);
	float alpha = G_FLOAT(OFS_PARM4);
	//Optional TODO2D: int flag = G_FLOAT(OFS_PARM5);
	/*TODO2D: size+rgb+alpha are discarded (size is pixels per char, rgb is 0-1), I wouldn't worry for software rendering anyway*/
	Draw_String(pos[0], pos[1], text);
}

/*loads the image if needed*/
void PFCS_drawpic (void)
{
	float *pos = G_VECTOR(OFS_PARM0);
	char *picname = G_STRING(OFS_PARM1);
	float *size = G_VECTOR(OFS_PARM2);
	float *rgb = G_VECTOR(OFS_PARM3);
	float alpha = G_FLOAT(OFS_PARM4);
	//Optional TODO2D: int flag = G_FLOAT(OFS_PARM5);
	qpic_t *pic;

	pic = Draw_CachePic(picname);
	if (!pic)
		return;

	/*TODO2D: size+rgb+alpha are discarded (size is screen pixels, rgb is 0-1), I wouldn't worry for software rendering anyway*/
	Draw_Pic(pos[0], pos[1], pic, DP_DEFAULT);
}

/*fills a box with a single colour*/
void PFCS_drawfill (void)
{
	float *pos = G_VECTOR(OFS_PARM0);
	float *size = G_VECTOR(OFS_PARM1);
	float *rgb = G_VECTOR(OFS_PARM2);
	float alpha = G_FLOAT(OFS_PARM3);

	/*TODO2D: RGB is forced to white=0xf*/
	Draw_Fill(pos[0], pos[1], size[0], size[1], 0xf, alpha);
}

/*an unnamed dp extension (dating back to menuqc), glScissor will do the job, optional extension so software rendering can be lazy*/
void PFCS_drawsetcliparea (void)
{
	/*Optional TODO2D*/
 	G_FLOAT(OFS_RETURN) = 0/*0=failed*/;
}
void PFCS_drawresetcliparea (void)
{
	/*Optional TODO2D*/
}

/*draws a string according to an engine's ^4 etc markup rules (engine doesn't need to support such rules, but does need to understand this function, basic q3 markup is expected, but not actually mandated anywhere)*/
void PFCS_drawcolorcodedstring (void)
{
	float *pos = G_VECTOR(OFS_PARM0);
	char *text = G_STRING(OFS_PARM1);
	float *size = G_VECTOR(OFS_PARM2);
	float alpha = G_FLOAT(OFS_PARM3);
	//Optional TODO: int flag = G_FLOAT(OFS_PARM4);

	/*TODO2D: size+alpha are discarded (size is pixels per char, alpha is 0-1), I wouldn't worry for software rendering anyway*/
	Draw_String(pos[0], pos[1], text);
}

/*return an ev_float stat*/
void PFCS_getstat_float (void)
{
	int stnum = G_FLOAT (OFS_PARM0);
	char *result = pr_progs->strings;
	float val;

	/*TODO: MAX_CL_STATS must be 128*/
	if (stnum == STAT_ITEMS)
		val = *(float *) &cl.items;
	else val = *(float *) &cl.stats[stnum];

	/*take an integer value, and cast it to a float*/
	G_FLOAT(OFS_RETURN) = val;
}

void PFCS_getstat_float_truncated (void)
{
	int stnum = G_FLOAT(OFS_PARM0);
	int first = G_FLOAT(OFS_PARM1);
	int count = G_FLOAT(OFS_PARM2);
	char *result = pr_progs->strings;

	unsigned int val;

	/*TODO: MAX_CL_STATS must be 128*/
	if (stnum == STAT_ITEMS)
		val = cl.items;
	else
		val = cl.stats[stnum];

	/*TODO: first+count are optional (defaulting to 0 and 32)*/

	if (first == 0 && count >= 32)
	{
		/*whole value, make sure its signed*/
		G_FLOAT(OFS_RETURN) = (int)val;
	}
	else
	{
		/*subsection of the value, reading a set of bits or so, before casting to a float, to grab the entire items stat*/
		G_FLOAT(OFS_RETURN) = (((unsigned int)val)&(((1<<count)-1)<<first))>>first;
	}
}

/*returns the string stored from a stat index. DP/H2 uses 4 adjacent stats for 16 chars, FTE uses an svc_statstring*/
void PFCS_getstat_string (void)
{
	int stnum = G_FLOAT (OFS_PARM0);
	char *result = PR_GetTempString ();

	/* TODO: use a protocol extension instead */
	((int *) result)[0] = cl.stats[stnum + 0];
	((int *) result)[1] = cl.stats[stnum + 1];
	((int *) result)[2] = cl.stats[stnum + 2];
	((int *) result)[3] = cl.stats[stnum + 3];
	result[16] = 0;

	G_INT(OFS_RETURN) = PR_SetString (result);
}

/*like setmodel except takes a modelindex instead of a string (sets modelindex and model still)*/
void PFCS_setmodelindex (void)
{
	/*TODO*/
}

/*returns the name of model in model precache slot X*/
void PFCS_modelnameforindex (void)
{
	/*TODO*/
}

/*provides an engine handle for a builtin particle effect, supported effects are engine specific, result is a handle(enum?) to a trail or point effect
0 = unsupported*/
void PFCS_particleeffectnum (void)
{
	/*TODO*/
	G_FLOAT(OFS_RETURN) = 0;
}

/*spawns the given particle effect as a trail between two points (may not work on point effects)*/
void PFCS_trailparticles (void)
{
	/*TODO*/
}

/*spawns the given particle effect around a single point (may not work on trail effects)*/
void PFCS_pointparticles (void)
{
	/*TODO*/
}

/*local centerprint, as if we received an svc_centerprint message (without passing it to the csqc handler)*/
void PFCS_centerprint (void)
{
	char *s = PF_VarString (0);
	SCR_CenterPrint(s);
}

/*local print - unconditionally calls Con_Printf("%s", str)*/
void PFCS_print (void)
{
	char *s = PF_VarString (0);

	Con_Printf("%s", s);
}

void PFCS_keynumtostring (void)
{
	/*TODO*/
}

void PFCS_stringtokeynum (void)
{
	/*TODO*/
}

void PFCS_getkeybind (void)
{
	/*TODO*/
}

//void PFCS_setcursormode (void){/*TODO*/}
//void PFCS_getmousepos (void){/*TODO*/}

/*engine keeps an input frame log, for prediction
this sets the input_foo globals to that input frame's various values*/
void PFCS_getinputstate (void)
{
	int f = G_FLOAT(OFS_PARM0);

	if (1/*TODO: verify that f is within the valid region of the log*/)
	{
		G_FLOAT(OFS_RETURN) = false;
	}
	else
	{
		/*TODO: set these globals to the log entry:
		input_timelength
		input_angles
		input_movevalues
		input_buttons
		input_impulse
		*/
		G_FLOAT(OFS_RETURN) = true;
	}
}

/*scales the user's sensitivity cvar in a volatile-safe way, useful for zoom but not much else*/
void PFCS_setsensitivityscaler (void)
{
	cl.qc.sensitivityscale = 1;
}

/*runs physics on self according to the input_foo globals (expected to be a player ent, includes a call to SV_ClientThink for every call)*/
void PFCS_runstandardplayerphysics (void)
{
	/*this is meant to be possible as a no-op*/
}

/*returns a string of info regarding a given player
negative player indexes give sorted players, by frags.
some specific keys exist. return an empty string for fields not known (eg: team).
like setinfo/userinfo in quakeworld, as far as a mod knows*/
void PFCS_getplayerkeyvalue (void)
{
	int pnum = G_FLOAT(OFS_PARM0);
	char *keyname = G_STRING(OFS_PARM1);
	int iret = 0;
	char *sret = NULL, *tmp;
	int rtype = 0;

	if (pnum < 0)
	{
		/*TODO: lookup frag sorting*/
	}
	if (pnum < 0 || pnum >= MAX_SCOREBOARD)
	{
		/*out of bounds*/
	}
	else if (!*cl.scores[pnum].name)
	{
	/*client not on server*/
	}
	else if (!strcmp(keyname, "name"))
	{
		rtype = 3; sret = cl.scores[pnum].name;
	}
	else if (!strcmp(keyname, "bottomcolor"))
	{
		rtype = 1; iret = cl.scores[pnum].colors & 0xf;
	}
	else if (!strcmp(keyname, "topcolor"))
	{
		rtype = 1; iret = (cl.scores[pnum].colors & 0xf0)>>4;
	}
	else if (!strcmp(keyname, "frags"))
	{
		rtype = 1; iret = cl.scores[pnum].frags;
	}
	else if (!strcmp(keyname, "ping"))
	{
		/*normally unknown in NQ*/
	}
	else if (!strcmp(keyname, "pl"))
	{
		/*normally unknown in NQ*/
	}
	else if (!strcmp(keyname, "entertime"))
	{
		rtype = 1; iret = cl.scores[pnum].entertime;
	}
	else if (!strcmp(keyname, "viewentity"))
	{
		rtype = 1; iret = pnum+1;
	}

	switch(rtype)
	{
	case 0:
		G_INT(OFS_RETURN) = 0;
		break;
	case 1:
		sret = PR_GetTempString ();
		sprintf(sret, "%i", iret);
		G_INT(OFS_RETURN) = PR_SetString(sret);
		break;
	case 2:
		tmp = PR_GetTempString ();
		strcpy(tmp, sret);
		G_INT(OFS_RETURN) = PR_SetString(tmp);
		break;
	case 3:
		G_INT(OFS_RETURN) = PR_SetString(sret);
		break;
	}
}
/*true if its playing a demo*/
void PFCS_isdemo (void)
{
	G_FLOAT(OFS_RETURN) = cls.demoplayback;
}

/*true if the engine is also the server (ie: if localcmd send to the server, and cvars always match)*/
void PFCS_isserver (void)
{
	G_FLOAT(OFS_RETURN) = sv.active;
}

/*sets the position and orientation of the audio code's listener, where you hear sound from*/
void PFCS_SetListener (void)
{
	/*TODO*/
}

/*registers a console command to trigger a call into qc*/

void PFCS_registercommand (void)
{
	char *cmdname = G_STRING(OFS_PARM0);

	Cbuf_AddText(va("\nalias \"%s\" cscq_command %s\n", cmdname, cmdname));
}
/*returns if an ent is free (only the current frame is reliable)*/
void PFCS_wasfreed (void)
{
	edict_t	*ent = (edict_t*)G_EDICT(OFS_PARM0);
	G_FLOAT(OFS_RETURN) = ent->free;
}

/*returns info about the server (like serverinfo in quakeworld)*/
void PFCS_serverkey (void)
{
	char *keyname = G_STRING(OFS_PARM0);
	char *result = pr_progs->strings;

	if (!strcmp (keyname, "ip"))
	{
		/*TODO*/
	}

	/*add extras here*/
	G_INT(OFS_RETURN) = PR_SetString (result);
}

/*should only be called from specific locations - csqc ents*/
void PFCS_ReadByte (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadByte ();
}
void PFCS_ReadChar (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadChar ();
}
void PFCS_ReadShort (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadShort ();
}
void PFCS_ReadLong (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadLong ();
}
void PFCS_ReadCoord (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadCoord(cl.protocol, cl.protocolflags);
}
void PFCS_ReadAngle (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadAngle(cl.protocol, cl.protocolflags);
}

void PFCS_ReadString (void)
{
	char *str = MSG_ReadString ();
	G_INT(OFS_RETURN) = PR_SetString (str);
}

void PFCS_ReadFloat (void)
{
	G_FLOAT(OFS_RETURN) = MSG_ReadFloat ();
}
void PFCS_ReadEntityNum (void)
{
	G_FLOAT(OFS_RETURN) = (unsigned short)MSG_ReadShort ();
}


ebfs_builtin_t cl_ebfs_builtins[] =
{
	{  0, NULL, PF_Fixme },		// has to be first entry as it is needed for initialization in PR_LoadProgs ()
	{  1, "makevectors", PF_makevectors},
	{  2, "setorigin", PF_setorigin},			/*TODO: SV_LinkEdict is server-only*/
	{  3, "setmodel", PF_setmodel},				/*TODO: SV_LinkEdict+indexes are server-only*/
	{  4, "setsize", PF_setsize},				/*TODO: SV_LinkEdict is server-only*/
	{  6, "break", PF_break},
	{  7, "random", PF_random},
	{  8, "sound", PF_sound},					/*TODO: this is server-only logic*/
	{  9, "normalize", PF_normalize},
	{ 10, "error", PF_error},
	{ 11, "objerror", PF_objerror},
	{ 12, "vlen", PF_vlen},
	{ 13, "vectoyaw", PF_vectoyaw},
	{ 14, "spawn", PF_Spawn},
	{ 15, "remove", PF_Remove},
	{ 16, "traceline", PF_traceline},			/*TODO: client collisions must trace against client ents, server ents may appear only as world*/
	{ 18, "find", PF_Find},
	{ 19, "precache_sound", PF_precache_sound},	/*TODO: client precaches shouldn't appear on the server. Also not a spawn function. Will break when acting as a client.*/
	{ 20, "precache_model", PF_precache_model},	/*TODO: client precaches shouldn't appear on the server. Also not a spawn function. Will break when acting as a client.*/
	{ 22, "findradius", PF_findradius},
	{ 26, "ftos", PF_ftos},
	{ 27, "vtos", PF_vtos},
	{ 28, "coredump", PF_coredump},
	{ 29, "traceon", PF_traceon},
	{ 30, "traceoff", PF_traceoff},
	{ 31, "eprint", PF_eprint},
	{ 32, "walkmove", PF_walkmove},				/*TODO: this is server-only logic*/
	{ 34, "droptofloor", PF_droptofloor},			/*TODO: this is server-only logic*/
	{ 35, "lightstyle", PF_lightstyle},			/*TODO: this is server-only logic*/
	{ 36, "rint", PF_rint},
	{ 37, "floor", PF_floor},
	{ 38, "ceil", PF_ceil},
	{ 40, "checkbottom", PF_checkbottom},			/*TODO: this is server-only logic*/
	{ 41, "pointcontents", PF_pointcontents},		/*TODO: this is server-only logic*/
	{ 43, "fabs", PF_fabs},
	{ 45, "cvar", PF_cvar},
	{ 46, "localcmd", PF_localcmd},
	{ 47, "nextent", PF_nextent},
	{ 48, "particle", PF_particle},				/*TODO: this is server-only logic*/
	{ 49, "ChangeYaw", PF_changeyaw},
	{ 51, "vectoangles", PF_vectoangles},
	{ 68, "precache_file", PF_precache_file},		/*no-op, safe*/
	{ 69, "makestatic", PF_makestatic},			/*TODO: this is server-only logic*/
	{ 72, "cvar_set", PF_cvar_set},
	{ 74, "ambientsound", PF_ambientsound},		/*TODO: this is server-only logic*/
	{ 75, "precache_model2", PF_precache_model},	/*TODO: this is server-only logic*/
	{ 76, "precache_sound2", PF_precache_sound},	/*TODO: this is server-only logic*/
	{ 77, "precache_file2", PF_precache_file},		/*no-op, safe*/
	{ 90, "tracebox", PF_tracebox },				/*TODO: this is server-only logic*/
	{ 99, "extension_find", PF_extension_find },	// 2001-10-20 Extension System by Lord Havoc/Maddes
	{ 114, "strlen", PF_strlen },
	{ 115, "strcat", PF_strcat },
	{ 116, "substring", PF_substring },
	{ 117, "stov", PF_stov },
	{ 118, "strzone", PF_strzone },
	{ 119, "strunzone", PF_strunzone },
	{222, "str2chr", PF_str2chr }, /*part of FTE_STRINGS*/
	{300, "R_ClearScene", PFCS_R_ClearScene},
	{301, "R_AddEntities", PFCS_R_AddEntities},
	{302, "R_AddEntity", PFCS_R_AddEntity},
	{303, "R_SetView", PFCS_R_SetView},
	{304, "R_RenderScene", PFCS_R_RenderScene},
	{305, "R_AddDynamicLight", PFCS_R_AddDynamicLight},
	{306, "R_BeginPolygon", PFCS_R_BeginPolygon},
	{307, "R_PolygonVertex", PFCS_R_PolygonVertex},
	{308, "R_EndPolygon", PFCS_R_EndPolygon},
	{310, "cs_unproject", PFCS_cs_unproject},
	{311, "cs_project", PFCS_cs_project},
	{315, "drawline", PFCS_drawline},
	{316, "iscachedpic", PFCS_iscachedpic},
	{317, "precache_pic", PFCS_precache_pic},
	{318, "draw_getimagesize", PFCS_draw_getimagesize},
	{319, "freepic", PFCS_freepic},
	{320, "drawcharacter", PFCS_drawcharacter},
	{321, "drawstring", PFCS_drawstring},
	{322, "drawpic", PFCS_drawpic},
	{323, "drawfill", PFCS_drawfill},
	{324, "drawsetcliparea", PFCS_drawsetcliparea},
	{325, "drawresetcliparea", PFCS_drawresetcliparea},
	{326, "drawcolorcodedstring", PFCS_drawcolorcodedstring},
	{330, "getstat_float", PFCS_getstat_float},
	{331, "getstat_float_truncated", PFCS_getstat_float_truncated},
	{332, "getstat_string", PFCS_getstat_string},
	{333, "setmodelindex", PFCS_setmodelindex},
	{334, "modelnameforindex", PFCS_modelnameforindex},
	{335, "particleeffectnum", PFCS_particleeffectnum},
	{336, "trailparticles", PFCS_trailparticles},
	{337, "pointparticles", PFCS_pointparticles},
	{338, "centerprint", PFCS_centerprint},
	{339, "print", PFCS_print},
	{340, "keynumtostring", PFCS_keynumtostring},
	{341, "stringtokeynum", PFCS_stringtokeynum},
	{342, "getkeybind", PFCS_getkeybind},
//	{343, "setcursormode", PFCS_setcursormode},
//	{344, "getmousepos", PFCS_getmousepos},
	{345, "getinputstate", PFCS_getinputstate},
	{346, "setsensitivityscaler", PFCS_setsensitivityscaler},
	{347, "runstandardplayerphysics", PFCS_runstandardplayerphysics},
	{348, "getplayerkeyvalue", PFCS_getplayerkeyvalue},
	{349, "isdemo", PFCS_isdemo},
	{350, "isserver", PFCS_isserver},
	{351, "SetListener", PFCS_SetListener},
	{352, "registercommand", PFCS_registercommand},
	{353, "wasfreed", PFCS_wasfreed},
	{354, "serverkey", PFCS_serverkey},
//	{359, "sendevent", PFCS_SendEvent},
	{360, "ReadByte", PFCS_ReadByte},
	{361, "ReadChar", PFCS_ReadChar},
	{362, "ReadShort", PFCS_ReadShort},
	{363, "ReadLong", PFCS_ReadLong},
	{364, "ReadCoord", PFCS_ReadCoord},
	{365, "ReadAngle", PFCS_ReadAngle},
	{366, "ReadString", PFCS_ReadString},
	{367, "ReadFloat", PFCS_ReadFloat},
	{368, "ReadEntitynum", PFCS_ReadEntityNum},
//	{369, "deltalisten", PFCS_DeltaListen},
};


void PR_InitEBFS (void)
{
	// init the EBFS tables
	sv_progs.ebfs_builtins = sv_ebfs_builtins;
	sv_progs.ebfs_numbuiltins = sizeof (sv_ebfs_builtins) / sizeof (ebfs_builtin_t);

	cl_progs.ebfs_builtins = cl_ebfs_builtins;
	cl_progs.ebfs_numbuiltins = sizeof (cl_ebfs_builtins) / sizeof (ebfs_builtin_t);
}


