
/*
This program is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; either version 3 of the License, or (at your option) any later 
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#define IQM_MAGIC "INTERQUAKEMODEL"
#define IQM_VERSION1 1	// this was a bug in the demo header file; 1 is the correct version
#define IQM_VERSION2 2	// this was a bug in the demo header file; 1 is the correct version

typedef struct iqmheader_s
{
    char magic[16];
    unsigned int version;
    unsigned int filesize;
    unsigned int flags;
    unsigned int num_text, ofs_text;
    unsigned int num_meshes, ofs_meshes;
    unsigned int num_vertexarrays, num_vertexes, ofs_vertexarrays;
    unsigned int num_triangles, ofs_triangles, ofs_adjacency;
    unsigned int num_joints, ofs_joints;
    unsigned int num_poses, ofs_poses;
    unsigned int num_anims, ofs_anims;
    unsigned int num_frames, num_framechannels, ofs_frames, ofs_bounds;
    unsigned int num_comment, ofs_comment;
    unsigned int num_extensions, ofs_extensions;
} iqmheader_t;

// standard MDL flags
#define	EF_ROCKET	1			// leave a trail
#define	EF_GRENADE	2			// leave a trail
#define	EF_GIB		4			// leave a trail
#define	EF_ROTATE	8			// rotate (bonus items)
#define	EF_TRACER	16			// green split trail
#define	EF_ZOMGIB	32			// small blood trail
#define	EF_TRACER2	64			// orange split trail + rotate
#define	EF_TRACER3	128			// purple trail


void IQM_Error (char *msg)
{
	printf ("ERROR: %s\n\n", msg);
	printf ("iqmflag <file> <options>\n");
	printf ("  each option is preceeded by '+' (add the flag) or '-' (remove the flag)\n");
	printf ("  options are applied in the order specified\n");
	printf ("\n");
	printf ("  rocket     leave a smoke trail\n");
	printf ("  grenade    leave a lighter smoke trail\n");
	printf ("  gib        leave a blood trail\n");
	printf ("  rotate     rotate (bonus item)\n");
	printf ("  tracer     green split trail\n");
	printf ("  zomgib     small blood trail\n");
	printf ("  tracer2    orange split trail + rotate\n");
	printf ("  tracer3    purple trail\n");
	printf ("  all        all flags\n");
}


int main (int argc, char *argv[])
{
	int i;
	int flags;
	FILE *f = NULL;
	iqmheader_t iqm;
	iqmheader_t *newfile;
	void *iqmbuf = NULL;

	if (argc < 3)
	{
		IQM_Error ("invalid number of arguments");
		return 0;
	}

	f = fopen (argv[1], "rb");

	if (!f)
	{
		IQM_Error ("could not open specified file");
		return 0;
	}

	fread (&iqm, sizeof (iqmheader_t), 1, f);

	if (ferror (f))
	{
		IQM_Error ("ferror on file read");
		fclose (f);
		return 0;
	}

	fclose (f);

	if (strcmp (iqm.magic, IQM_MAGIC))
	{
		IQM_Error ("specified file is not an IQM file");
		return 0;
	}

	if (!(iqm.version == IQM_VERSION1 || iqm.version == IQM_VERSION2))
	{
		IQM_Error ("IQM file has bad version (should be 1 or 2)");
		return 0;
	}

	iqmbuf = malloc (iqm.filesize);
	newfile = (iqmheader_t *) iqmbuf;

	f = fopen (argv[1], "rb");
	fread (iqmbuf, iqm.filesize, 1, f);
	fclose (f);

	// store out initial flags that we'll modify
	flags = iqm.flags;

	for (i = 2; i < argc; i++)
	{
		int flag = 0;

		if (!strcmp (&argv[i][1], "rocket"))
			flag = EF_ROCKET;
		else if (!strcmp (&argv[i][1], "grenade"))
			flag = EF_GRENADE;
		else if (!strcmp (&argv[i][1], "gib"))
			flag = EF_GIB;
		else if (!strcmp (&argv[i][1], "rotate"))
			flag = EF_ROTATE;
		else if (!strcmp (&argv[i][1], "tracer"))
			flag = EF_TRACER;
		else if (!strcmp (&argv[i][1], "zomgib"))
			flag = EF_ZOMGIB;
		else if (!strcmp (&argv[i][1], "tracer2"))
			flag = EF_TRACER2;
		else if (!strcmp (&argv[i][1], "tracer3"))
			flag = EF_TRACER3;
		else if (!strcmp (&argv[i][1], "all"))
			flag = 0xff;
		else
		{
			printf ("unknown flag: \"%s\"\n", argv[i]);
			continue;
		}

		if (argv[i][0] == '+') flags |= flag;
		if (argv[i][0] == '-') flags &= ~flag;
	}

	newfile->flags = flags & 0xff;

	f = fopen (argv[1], "wb");
	fwrite (iqmbuf, iqm.filesize, 1, f);
	fclose (f);

	free (iqmbuf);

	return 0;
}
